#include "stdafx.h"
#define FUNCTION_PROFILE
#include "glb.h"

#ifdef min
#undef min
#undef max
#endif

#include <list>
#include <map>
#include <string>

/* profiler record: registers at function entry and exit points */
struct ProfRecord
{
	bool			m_bEntry;
	LARGE_INTEGER	m_lnTimeStamp;  // "total-time" timestamp
	LARGE_INTEGER	m_lnTimeStampE; // "execution" timestamp
	const char*			m_pszFuncName;
};

/* function record: records the statistics for each function */
struct FuncRecord
{
	int		m_nInvoke;
	int		m_nTotElapsed;
	int		m_nExecElapsed;
};

/* helper class: a stack for calculation of duration of each invocation of fuction */
class GFProfilerStack
{
public:
	GFProfilerStack(){}

	~GFProfilerStack(){}

	void Push(ProfRecord* lpProfRec);
	ProfRecord* Pop();

	// AddStack() increase the entry time to offset other function execution times
	void AddStack(ULONGLONG ullTickDiff);

	int GetCount();

private:
	std::list<ProfRecord*> m_listProfRec;
};


std::list<ProfRecord*>& GetFullList()
{
static std::list<ProfRecord*> g_listProfRec; // a list if ProfRecord
	return g_listProfRec;
}


GFProfilerClass::GFProfilerClass(const char* lpszFuncName)
{
	ProfRecord* pProfRec = new ProfRecord;
	pProfRec->m_bEntry = TRUE;
	::QueryPerformanceCounter(&(pProfRec->m_lnTimeStamp));
	pProfRec->m_lnTimeStampE = pProfRec->m_lnTimeStamp; // initialize the "execution time" timestamp
	pProfRec->m_pszFuncName = lpszFuncName;
	m_pszFuncName = lpszFuncName;
	GetFullList().push_back(pProfRec);
}

GFProfilerClass::~GFProfilerClass()
{
	ProfRecord* pProfRec = new ProfRecord;
	pProfRec->m_bEntry = FALSE;
	::QueryPerformanceCounter(&(pProfRec->m_lnTimeStamp));
	pProfRec->m_pszFuncName = m_pszFuncName;
	GetFullList().push_back(pProfRec);
}

void GFProfilerClass::ExportProf(const char* cstrFileName, const char* date, const char* time)
{
std::string strFileName(cstrFileName);
GFProfilerStack stackProfRec; // a stack for calculation of the execution times
std::map<const char*, FuncRecord*> g_mapFuncRec; // a table of statistics

	FILE* file;
	if(file= fopen(strFileName.c_str(), "wb"))
	{
		// calculate the CPU frequency
		LARGE_INTEGER liQPFreq;
		QueryPerformanceFrequency(&liQPFreq);

		ULONGLONG llQPFreq; // the CPU frequency
		llQPFreq = liQPFreq.QuadPart/1000; // unit of sec

		// application info
		int freq = (int)llQPFreq;
		fprintf(file, "Date,Time,QuerPerfFreq\n");
		fprintf(file, "%s,%s,%d\n", date,time,freq);

		// iterate the function entry/exit records
		// record the nb of invocation, durations, etc into a map with function names as keys
		for (std::list<ProfRecord*>::iterator pos = GetFullList().begin(); pos!=GetFullList().end(); ++pos)
		{
			ProfRecord* pProfRec = *pos;

			if (pProfRec->m_bEntry) // function entry
			{
				stackProfRec.Push(pProfRec);
			}
			else  // function exit
			{
				ProfRecord* pProfRecEntry = stackProfRec.Pop();

				if (pProfRecEntry)
				{
					ULONGLONG ullTickDiff = pProfRec->m_lnTimeStamp.QuadPart - pProfRecEntry->m_lnTimeStamp.QuadPart;

					// adjust for execution time
					// if functions Foo() and Bar() are being profiled and Foo() calls Bar(),
					// the execution time of Bar() should be deducted from Foo()'s
					// the get the "actual" time elapsed in Foo()
					ULONGLONG ullTickDiffE = pProfRec->m_lnTimeStamp.QuadPart - pProfRecEntry->m_lnTimeStampE.QuadPart;
					stackProfRec.AddStack(ullTickDiffE); 
					
					// keep statistics in a function-statistics map
					int nDuration = (int)((ullTickDiff) / llQPFreq);
					int nDurationE = (int)((ullTickDiffE) / llQPFreq);

					std::map<const char*, FuncRecord*>::iterator it_fr;
					it_fr = g_mapFuncRec.find(pProfRec->m_pszFuncName);
					if (it_fr!=g_mapFuncRec.end()) // existing function, add statistics
					{
						it_fr->second->m_nInvoke++;
						it_fr->second->m_nTotElapsed += nDuration;
						it_fr->second->m_nExecElapsed += nDurationE;
					}
					else // new function, initialize statistics
					{
						FuncRecord* pFuncRec = new FuncRecord;
						g_mapFuncRec[pProfRec->m_pszFuncName] = pFuncRec;
						pFuncRec->m_nInvoke = 1;
						pFuncRec->m_nTotElapsed = nDuration;
						pFuncRec->m_nExecElapsed = nDurationE;
					}
				}
			}
		}
		
		// "close" still open functions (d'tor not called, yet)
		ProfRecord QuitRec;
		QuitRec.m_bEntry = FALSE;
		::QueryPerformanceCounter(&(QuitRec.m_lnTimeStamp));

		ProfRecord* pProfRecEntry;
		while(pProfRecEntry = stackProfRec.Pop())
		{
			ULONGLONG ullTickDiff = QuitRec.m_lnTimeStamp.QuadPart - pProfRecEntry->m_lnTimeStamp.QuadPart;

			// adjust for execution time
			// if functions Foo() and Bar() are being profiled and Foo() calls Bar(),
			// the execution time of Bar() should be deducted from Foo()'s
			// the get the "actual" time elapsed in Foo()
			ULONGLONG ullTickDiffE = QuitRec.m_lnTimeStamp.QuadPart - pProfRecEntry->m_lnTimeStampE.QuadPart;
			stackProfRec.AddStack(ullTickDiffE); 
			
			// keep statistics in a function-statistics map
			int nDuration = (int)((ullTickDiff) / llQPFreq);
			int nDurationE = (int)((ullTickDiffE) / llQPFreq);

			std::map<const char*, FuncRecord*>::iterator it_fr;
			it_fr = g_mapFuncRec.find(pProfRecEntry->m_pszFuncName);
			if (it_fr!=g_mapFuncRec.end()) // existing function, add statistics
			{
				it_fr->second->m_nInvoke++;
				it_fr->second->m_nTotElapsed += nDuration;
				it_fr->second->m_nExecElapsed += nDurationE;
			}
			else // new function, initialize statistics
			{
				FuncRecord* pFuncRec = new FuncRecord;
				g_mapFuncRec[pProfRecEntry->m_pszFuncName] = pFuncRec;
				pFuncRec->m_nInvoke = 1;
				pFuncRec->m_nTotElapsed = nDuration;
				pFuncRec->m_nExecElapsed = nDurationE;
			}
		}
		

		FuncRecord TotRec = { 0, 0, 0 }; // totals

		// writing the results into a CSV file
		fprintf(file, "Function,Calls,Tot.with childs[s],Exec.only[s],Av.Exe[s/call]\n");

		for (std::map<const char*, FuncRecord*>::iterator pos = g_mapFuncRec.begin(); pos != g_mapFuncRec.end(); ++pos)
		{
			FuncRecord* pFuncRec=pos->second;
			TotRec.m_nInvoke += pFuncRec->m_nInvoke;
			TotRec.m_nExecElapsed += pFuncRec->m_nExecElapsed;
			TotRec.m_nTotElapsed += pFuncRec->m_nTotElapsed;
			
			double dbTotElapsed = pFuncRec->m_nTotElapsed;
			double dbExecElapsed = pFuncRec->m_nExecElapsed;
			fprintf(file,
				"%s,%d,%.3lf,%.3lf,%.3lf\n", 
				pos->first,
				pFuncRec->m_nInvoke,
				dbTotElapsed/1000.0,
				dbExecElapsed/1000.0,
				pFuncRec->m_nInvoke ? (dbExecElapsed/1000.0 / pFuncRec->m_nInvoke) : 0
				);
		}

		fprintf(file, "%s,%d,%.3lf,%.3lf\n", 
			"Total", 
			TotRec.m_nInvoke, 
			(double)(TotRec.m_nTotElapsed/1000.0 ), 
			(double)(TotRec.m_nExecElapsed/1000.0 ));
		fclose(file);
	}
	else
	{
		// Unable to open '%s' for writing the profiling results!
	}

	// free up the memory in the collections
	for (std::map<const char*, FuncRecord*>::iterator pos = g_mapFuncRec.begin(); pos != g_mapFuncRec.end(); ++pos)
	{
		FuncRecord* pFuncRec=pos->second;
		delete pFuncRec;
	}
	g_mapFuncRec.clear();

	for (std::list<ProfRecord*>::iterator pos = GetFullList().begin(); pos!=GetFullList().end(); ++pos)
	{
		ProfRecord* pProfRec = *pos;
		if (pProfRec)
			delete pProfRec;
	}
	GetFullList().clear();
}



void GFProfilerStack::Push(ProfRecord* lpProfRec)
{
	m_listProfRec.push_front(lpProfRec);
}
ProfRecord* GFProfilerStack::Pop()
{
	if (m_listProfRec.size() > 0)
	{
		ProfRecord* pP = m_listProfRec.front();
		m_listProfRec.pop_front();
		return pP;
	}
	else
		return NULL;
}
void GFProfilerStack::AddStack(ULONGLONG ullTickDiff)
{
	for (std::list<ProfRecord*>::iterator pos = m_listProfRec.begin(); pos!=m_listProfRec.end(); ++pos)
	{
		ProfRecord* pProfRec = *pos;
		if (pProfRec)
			pProfRec->m_lnTimeStampE.QuadPart += ullTickDiff;
	}
}

int GFProfilerStack::GetCount()
{
	return (int)m_listProfRec.size();
}

