/*----------------------------------------------------------------------------
  astarplanner.cpp
  Implementation for AStarPlanner class.

  Copyright (C) 2003 Ted Cipicchio <ted@thereisnospork.com>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to
  deal in the Software without restriction, including without limitation the
  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  IN THE SOFTWARE.
----------------------------------------------------------------------------*/

#include "stdafx.h"
#include "astarplanner.h"

// Compute heuristic weight of node based on rough distance calculation.
#define COMPUTE_DIST(CURR_X, CURR_Y, DEST_X, DEST_Y) \
	(float(abs((DEST_X) - (CURR_X)) + abs((DEST_Y) - (CURR_Y))))


// Check the specified node against the open and closed lists and add to open.
#define CHECK_ADD_PRE(X, Y, MAP_IDX)  \
	do                                                           \
	{                                                            \
		/* Make sure we're not off the map. */                   \
		if ((X) < 0 || (X) >= (int)width || (Y) < 0 || (Y) >= (int)height) \
			break;                                               \
		                                                         \
		/* Check for wall. */                                    \
		float map_val = themap[MAP_IDX];                         \
		if (map_val == PLANNER_WALL)                             \
			break;


#define CHECK_ADD_POST(X, Y) \
		/* Check closed list. */                                 \
		Node *ptest = closed_iter.find(X, Y);                    \
		if (ptest)                                               \
		{                                                        \
			/* Check if found node is better than our node. */   \
			Node &rtest = *ptest;                                \
			if (rtest.cost <= cost)                              \
				break;                                           \
			                                                     \
			/* This node is better, so replace the other. */     \
			closed_iter.remove();                                \
			rtest.pParent = pnode;                               \
			rtest.x = (X);                                       \
			rtest.y = (Y);                                       \
			rtest.cost = cost;                                   \
			float heuristic = COMPUTE_DIST(X, Y, end_x, end_y);  \
			rtest.weight = (heuristic - cost) * heuristic_weight \
				+ cost;                                          \
			SortedInsert(open, ptest);                           \
			                                                     \
			break;                                               \
		}                                                        \
		                                                         \
		/* Check open list. */                                   \
		unsigned int i = open.size();                            \
		bool bstop = false;                                      \
		while (i--)                                              \
		{                                                        \
			Node &rcurr = *open[i];                              \
			if ((X) != rcurr.x || (Y) != rcurr.y)                \
				continue;                                        \
			                                                     \
			/* Gonna stop either way... */                       \
			bstop = true;                                        \
			                                                     \
			/* Found a node, is it better? */                    \
			if (rcurr.cost <= cost)                              \
				break;                                           \
			                                                     \
			/* This node is better, so replace the other. */     \
			open.remove(i);                                      \
			rcurr.pParent = pnode;                               \
			rcurr.x = (X);                                       \
			rcurr.y = (Y);                                       \
			rcurr.cost = cost;                                   \
			float heuristic = COMPUTE_DIST(X, Y, end_x, end_y);  \
			rcurr.weight = (heuristic - cost) * heuristic_weight \
				+ cost;                                          \
			SortedInsert(open, &rcurr);                          \
			                                                     \
			break;                                               \
		}                                                        \
		if (bstop)                                               \
			break;                                               \
		                                                         \
		/* Not in either list, so it should be added. */         \
		Node *pchild = factory.fetch();                          \
		Node &rchild = *pchild;                                  \
		rchild.pParent = pnode;                                  \
		rchild.x = (X);                                          \
		rchild.y = (Y);                                          \
		rchild.cost = cost;                                      \
	    float heuristic = COMPUTE_DIST(X, Y, end_x, end_y);      \
		rchild.weight = (heuristic - cost) * heuristic_weight    \
			+ cost;                                              \
		SortedInsert(open, pchild);                              \
	} while(false);

#define CHECK_ADD_NORM(X, Y, MAP_IDX)     \
	CHECK_ADD_PRE(X, Y, MAP_IDX)          \
		float cost = node_cost + map_val; \
	CHECK_ADD_POST(X, Y)

#define CHECK_ADD_DIAG(X, Y, MAP_IDX)                 \
	if(mt_description.can_walk_diagonally){           \
	CHECK_ADD_PRE(X, Y, MAP_IDX)                      \
		float cost = node_cost + map_val * diag_cost; \
	CHECK_ADD_POST(X, Y)  }

static void SortedInsert (FixedDeque<Node *> &nodeDeque, Node *pNode)
{
	unsigned int lo = 0;
	unsigned int hi = nodeDeque.size();
	while (lo < hi)
	{
		unsigned int mid = (lo + hi) >> 1;  // (lo + hi) / 2
		Node *ptest = nodeDeque[mid];
		if (pNode->weight < ptest->weight)
		{
			lo = mid + 1;
			continue;
		}
		hi = mid;
	}
//	ASSERT(lo == hi);
	nodeDeque.insert(lo, pNode);
}


//----------------------------------------------------------------------------
// Public Methods
//----------------------------------------------------------------------------

// Initialize the path-finding algorithm.
void AStarPlanner::initialize ()
{
	// Base class initialize()...
	Planner::initialize();

	// Get node for starting point and put into queue.
	Node *pnode = mt_nodeFactory.fetch();
	pnode->pParent = 0;
	pnode->x = mt_description.startX;
	pnode->y = mt_description.startY;
	pnode->cost = 0;
	mt_open.pushBack(pnode);
}


#if 0
	bool AStarPlanner::run ()
	{
	return false;
	}

#else
	// Run a single time slice of the path-finding algorithm.
	bool AStarPlanner::run ()
	{
		// Increment run count.
		mt_numSlices++;

		// Get some useful variables.
		int end_x = mt_description.endX;
		int end_y = mt_description.endY;
		unsigned int width  = mt_description.width;
		unsigned int height = mt_description.height;
		float *themap = mt_map;
		FixedDeque<Node *> &open = mt_open;
		NodeHashTable &closed = mt_closed;
		NodeHashIterator closed_iter(&closed);
		NodeFactory &factory = mt_nodeFactory;
	//	unsigned int slice_time = 0; //mt_description.sliceTime;
		float diag_cost = (mt_description.bDiagonalCost ? 1.4142136f : 1.0f);
		float heuristic_weight = mt_description.heuristicWeight;

		// Create a timer and start it.
	//	Timer timer; timer.start();

		// Cycle through open nodes.
		while (!open.isEmpty())
		{
			// Get the next node to examine.
			Node *pnode = open.back();
			open.popBack();
			Node &rnode = *pnode;
			int x = rnode.x;
			int y = rnode.y;

			// Check if the node is the end node.
			if (x == end_x && y == end_y)
			{
				closed.add(pnode);
				mt_pGoal = pnode;
	//			timer.stop();
	//			mt_runningTime += timer.get();
				mt_bDone = true;
				return true;
			}

			// More information...
			int map_idx = x + y * width;
			float node_cost = rnode.cost;

			// Add this node's children to the open list.
			int child_x = x - 1;
			int child_y = y - 1;

#if !(defined(POCKETPC) && defined(_DEBUG))
			CHECK_ADD_DIAG(child_x, child_y, map_idx - width - 1)
			child_x++;
			CHECK_ADD_NORM(child_x, child_y, map_idx - width)
			child_x++;
			CHECK_ADD_DIAG(child_x, child_y, map_idx - width + 1)
			child_y++;
			CHECK_ADD_NORM(child_x, child_y, map_idx + 1)
			child_x -= 2;
			CHECK_ADD_NORM(child_x, child_y, map_idx - 1)
			child_y++;
			CHECK_ADD_DIAG(child_x, child_y, map_idx + width - 1)
			child_x++;
			CHECK_ADD_NORM(child_x, child_y, map_idx + width)
			child_x++;
			CHECK_ADD_DIAG(child_x, child_y, map_idx + width + 1)
#endif

			// Add the node to the closed list.
			closed.add(pnode);

			// Check if we've run out of time for this slice.
	//		if (slice_time && timer.get() > slice_time) break;
		}

	//	timer.stop();
	//	mt_runningTime += timer.get();
		mt_bDone = open.isEmpty();
		return mt_bDone;
	}

#endif
