#ifndef FIXEDDEQUE_H
#define FIXEDDEQUE_H
/**
  @class FixedDeque
  @brief Fixed-size deque class.

  This is a lightweight template deque (double-ended queue) of fixed size,
  built primarily for speed improvements (doesn't reallocate memory unless
  explicitly told to do so).  Like the STL deque, it supports constant time
  insertion of entries at the beginning and end and linear time insertion of
  entries within the deque.  Unlike the STL vector, memory is not reallocated
  once the deque gets full.

  Copyright (C) 2003 Ted Cipicchio <ted@thereisnospork.com>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to
  deal in the Software without restriction, including without limitation the
  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  IN THE SOFTWARE.

  @date  Created: Mon Mar 24 2003
*/

//----------------------------------------------------------------------------
template <typename T>
class FixedDeque
{
public:
	//------------------------------------------------------------------------
	/** @name Construction/Destruction */
	//@{

	/// Constructor.
	FixedDeque (unsigned int maxSize = 0);
	/// Copy constructor.
	FixedDeque (const FixedDeque &src);
	/// Destructor.
	virtual ~FixedDeque ();

	//@}


	//------------------------------------------------------------------------
	/** @name General Operations */
	//@{

	/// Allocate memory for the deque.
	bool allocate (unsigned int maxSize);

	/// Add an entry to the beginning of the deque.
	bool pushFront (const T &data);
	/// Add an entry to the end of the deque.
	bool pushBack (const T &data);
	/// Insert an entry into the specified position.
	bool insert (unsigned int pos, const T &data);

	/// Remove the entry at the beginning of the deque.
	bool popFront ();
	/// Remove the entry at the end of the deque.
	bool popBack ();
	/// Remove an entry from the specified position.
	bool remove (unsigned int pos);

	/// Clear out the deque.
	void clear ();

	/// Get the data at the front of the deque.
	T &front ();
	/// Get the data at the front of the deque (constant).
	const T &front () const;

	/// Get the data at the back of the deque.
	T &back ();
	/// Get the data at the back of the deque (constant).
	const T &back () const;

	/// Get the current size of the deque.
	unsigned int size () const;
	/// Get the maximum size of the deque.
	unsigned int capacity () const;

	/// Check if the deque is empty.
	bool isEmpty () const;
	/// Check if the deque is full.
	bool isFull () const;

	//@}


	//------------------------------------------------------------------------
	/** @name Overloaded Operators */
	//@{

	/// Assignment operator.
	FixedDeque &operator = (const FixedDeque &src);

	/// Array access.
	T &operator [] (unsigned int idx);
	/// Array access (constant).
	const T &operator [] (unsigned int idx) const;
	//@}

private:
	/// Array of data.
	T *mv_data;
	/// Secondary data buffer.
	T *mv_dataSecond;
	/// Size of the array (maximum number of possible entries).
	unsigned int mv_size;

	/// Index of first element.
	unsigned int mv_firstIdx;
	/// Index of element after last.
	unsigned int mv_lastIdx;
	/// True if the array is empty (needed in special cases).
	bool mv_bEmpty;


	/// Set this object's data based on data in another object.
	void v_set (const FixedDeque &src);
	/// Free allocated memory.
	void v_free ();
};


//----------------------------------------------------------------------------
// Construction/Destruction
//----------------------------------------------------------------------------

/// Constructor.
/**
  @param maxSize  Size to initially allocate the deque to. (default = 0)

  @note  The deque can also be allocated using allocate().
*/
template <typename T>
FixedDeque<T>::FixedDeque (unsigned int maxSize)
	: mv_size(maxSize), mv_firstIdx(0), mv_lastIdx(0), mv_bEmpty(true)
{
	// Allocate the specified memory for the data.
	if (maxSize)
	{
		mv_data = new T [maxSize];
		mv_dataSecond = new T [maxSize];
	}
	else
	{
		mv_data = 0;
		mv_dataSecond = 0;
	}
}


/// Copy constructor.
/**
  @param src  Source object to make a copy from.
*/
template <typename T>
FixedDeque<T>::FixedDeque (const FixedDeque &src)
{
	// Set the object's data.
	v_set(src);
}


/// Destructor.
template <typename T>
FixedDeque<T>::~FixedDeque ()
{
	// Free up memory.
	v_free();
}


//----------------------------------------------------------------------------
// Public Methods
//----------------------------------------------------------------------------

/// Allocate memory for the deque.
/**
  @param maxSize  Size to allocate in the deque.

  @return  True if successful, false if not.
*/
template <typename T>
bool FixedDeque<T>::allocate (unsigned int maxSize)
{
	// Clear out any previous deque array.
	if (mv_data)
	{
		delete [] mv_data;
		delete [] mv_dataSecond;
	}

	// Reset data and set size.
	mv_size = maxSize;
	mv_firstIdx = mv_lastIdx = 0;
	mv_bEmpty = true;

	// Allocate memory as necessary.
	if (maxSize)
	{
		mv_data = new T [maxSize];
		mv_dataSecond = new T [maxSize];
	}
	else
	{
		mv_data = 0;
		mv_dataSecond = 0;
	}

	return true;
}


/// Add an entry to the beginning of the deque.
/**
  @param data  Data to add.

  @return  True if added successfully, false if not.

  @see popFront(), pushBack(), insert()
*/
template <typename T>
bool FixedDeque<T>::pushFront (const T &data)
{
	// Check first if the deque is full.
	if (mv_firstIdx == mv_lastIdx && !mv_bEmpty)
		return false;

	// Adjust the start index, taking care to wrap value as necessary.
	mv_firstIdx--;
	if (mv_firstIdx >= mv_size)
		mv_firstIdx = mv_size - 1;

	// Add the data to the deque.
	mv_data[mv_firstIdx] = data;
	mv_bEmpty = false;

	return true;
}


/// Add an entry to the end of the deque.
/**
  @param data  Data to add.

  @return  True if added successfully, false if not.

  @see popBack(), pushFront(), insert()
*/
template <typename T>
bool FixedDeque<T>::pushBack (const T &data)
{
	// Check first if the deque is full.
	if (mv_firstIdx == mv_lastIdx && !mv_bEmpty)
		return false;

	// Add the data to the deque.
	mv_data[mv_lastIdx] = data;
	mv_bEmpty = false;

	// Adjust the end index, taking care to wrap value as necessary.
	mv_lastIdx++;
	if (mv_lastIdx >= mv_size)
		mv_lastIdx = 0;

	return true;
}


/// Insert an entry into the specified position.
/**
  @param pos   Index of position to add data to (data will be inserted
               directly before the data currently there).
  @param data  Data to add.

  @return  True if added successfully, false if not.

  @note  @pos can be equal to the current size of the deque (one past the last
         index).  This is essentially the same as adding the entry to the end
         using pushBack(), just slower...

  @see remove(), pushFront(), pushBack()
*/
template <typename T>
bool FixedDeque<T>::insert (unsigned int pos, const T &data)
{
	// Special cases...
	unsigned int pos_mod = mv_firstIdx + pos;
	if (pos_mod >= mv_size)
		pos_mod -= mv_size;
	if (pos_mod == mv_lastIdx)
		return pushBack(data);
	if (pos_mod == mv_firstIdx)
		return pushFront(data);

	// Check first if the deque is full.
	if (mv_firstIdx == mv_lastIdx && !mv_bEmpty)
		return false;

	// Determine the maximum index and check if the index given is less than
	// or equal to it.
	unsigned int max_idx = (mv_lastIdx >= mv_firstIdx
		? mv_lastIdx - mv_firstIdx
		: mv_size - mv_firstIdx + mv_lastIdx);
	if (pos > max_idx)
		return false;

	// Data will be added, so unset empty flag.
	mv_bEmpty = false;

	unsigned int first_idx = mv_firstIdx;
	unsigned int pos_idx = first_idx + pos;
	unsigned int last_idx = (mv_lastIdx ? mv_lastIdx - 1 : mv_size - 1);
	unsigned int first_dst = 0;
	unsigned int last_dst = max_idx;
	if (pos_idx >= mv_size)
	{
		// Data wraps between first and position elements, so copy block of
		// first up to wrap and change first destination index.
		pos_idx -= mv_size;
		unsigned int diff = mv_size - first_idx;
		memcpy(&mv_dataSecond[0], &mv_data[first_idx], sizeof(T) * diff);
		first_idx = 0;
		first_dst = diff;
	}
	else if (last_idx < pos_idx)
	{
		// Data wraps between position and last elements, so copy block of
		// array start up to last and change last destination index.
		unsigned int dest = last_dst - last_idx;
		memcpy(&mv_dataSecond[dest], &mv_data[0],
			sizeof(T) * (last_idx + 1));
		last_idx = mv_size - 1;
		last_dst = dest - 1;
	}
	// Copy rest of data over.
	memcpy(&mv_dataSecond[first_dst], &mv_data[first_idx],
		sizeof(T) * (pos - first_dst));
	mv_dataSecond[pos] = data;
	memcpy(&mv_dataSecond[pos + 1], &mv_data[pos_idx],
		sizeof(T) * (last_dst - pos));

	// Swap pointers.
	T *ptmp = mv_data;
	mv_data = mv_dataSecond;
	mv_dataSecond = ptmp;

	// Set new indeces.
	mv_firstIdx = 0;
	mv_lastIdx = max_idx + 1;

	return true;
}


/// Remove the entry at the beginning of the deque.
/**
  @return  True if removed successfully, false if not.

  @note  The data isn't actually deleted, as the array is preallocated.  If
         the entry has any memory allocated you want to free, member functions
         of the object should be called prior to this function to free up the
         memory.

  @see pushFront(), popBack(), remove(), clear()
*/
template <typename T>
bool FixedDeque<T>::popFront ()
{
	// Check first if the deque is empty.
	if (mv_bEmpty)
		return false;

	// Adjust the start index, taking care to wrap value as necessary.
	mv_firstIdx++;
	if (mv_firstIdx >= mv_size)
		mv_firstIdx = 0;

	// Check if the deque is now empty.
	if (mv_firstIdx == mv_lastIdx)
		mv_bEmpty = true;

	return true;
}


/// Remove the entry at the end of the deque.
/**
  @return  True if removed successfully, false if not.

  @note  The data isn't actually deleted, as the array is preallocated.  If
         the entry has any memory allocated you want to free, member functions
         of the object should be called prior to this function to free up the
         memory.

  @see pushBack(), popFront(), remove(), clear()
*/
template <typename T>
bool FixedDeque<T>::popBack ()
{
	// Check first if the deque is empty.
	if (mv_bEmpty)
		return false;

	// Adjust the end index, taking care to wrap value as necessary.
	mv_lastIdx--;
	if (mv_lastIdx >= mv_size)
		mv_lastIdx = mv_size - 1;

	// Check if the deque is now empty.
	if (mv_firstIdx == mv_lastIdx)
		mv_bEmpty = true;

	return true;
}


/// Remove an entry from the specified position.
/**
  @param pos  Index of the entry to remove.

  @return  True if removed successfully, false if not.

  @note  The data isn't actually deleted, as the array is preallocated.  If
         the entry has any memory allocated you want to free, member functions
         of the object should be called prior to this function to free up the
         memory.

  @see insert(), popFront(), popBack(), clear()
*/
template <typename T>
bool FixedDeque<T>::remove (unsigned int pos)
{
	// Special cases...
	// Note: This also avoids the code below from blowing up if the deque only
	// has one or two elements.
	unsigned int pos_mod = mv_firstIdx + pos;
	if (pos_mod >= mv_size)
		pos_mod -= mv_size;
	if (pos_mod == (mv_lastIdx ? mv_size - 1 : mv_lastIdx - 1))
		return popBack();
	if (pos_mod == mv_firstIdx)
		return popFront();

	// Check first if the deque is empty.
	if (mv_bEmpty)
		return false;

	// Determine the maximum index and check if the index given is less than
	// it.
	// Note that this is somewhat different than in insert() since the deque
	// can be full but not empty (instead of can be empty but not full).
	unsigned int max_idx = (mv_lastIdx > mv_firstIdx
		? mv_lastIdx - mv_firstIdx
		: mv_size - mv_firstIdx + mv_lastIdx);
	if (pos >= max_idx)
		return false;

	unsigned int first_idx = mv_firstIdx;
	unsigned int pos_idx = first_idx + pos;
	unsigned int last_idx = (mv_lastIdx ? mv_lastIdx - 1 : mv_size - 1);
	unsigned int first_dst = 0;
	unsigned int last_dst = max_idx - 2;
	if (pos_idx >= mv_size)
	{
		// Data wraps between first and position elements, so copy block of
		// first up to wrap and change first destination index.
		pos_idx -= mv_size;
		unsigned int diff = mv_size - first_idx;
		memcpy(&mv_dataSecond[0], &mv_data[first_idx], sizeof(T) * diff);
		first_idx = 0;
		first_dst = diff;
	}
	else if (last_idx < pos_idx)
	{
		// Data wraps between position and last elements, so copy block of
		// array start up to last and change last destination index.
		unsigned int dest = last_dst - last_idx;
		memcpy(&mv_dataSecond[dest], &mv_data[0],
			sizeof(T) * (last_idx + 1));
		last_idx = mv_size - 1;
		last_dst = dest - 1;
	}
	// Copy rest of data over.
	memcpy(&mv_dataSecond[first_dst], &mv_data[first_idx],
		sizeof(T) * (pos - first_dst));
	memcpy(&mv_dataSecond[pos], &mv_data[pos_idx + 1],
		sizeof(T) * (last_dst - pos + 1));

	// Swap pointers.
	T *ptmp = mv_data;
	mv_data = mv_dataSecond;
	mv_dataSecond = ptmp;

	// Set new indeces.
	mv_firstIdx = 0;
	mv_lastIdx = max_idx - 1;

	// Check if the deque is now empty.
	if (!mv_lastIdx)
		mv_bEmpty = true;

	return true;
}


/// Clear out the deque.
/**
  This resets deque indeces to give an empty deque.

  @note  The data isn't actually deleted, as the array is preallocated.  If
         the entry has any memory allocated you want to free, member functions
         of the object should be called prior to this function to free up the
         memory.

  @see popFront(), popBack(), remove()
*/
template <typename T>
void FixedDeque<T>::clear ()
{
	mv_firstIdx = mv_lastIdx = 0;
	mv_bEmpty = true;
}


/// Get the data at the front of the deque.
/**
  @return  Reference to the data at the front of the deque.

  @note  Calling this on an empty deque has unexpected results.
*/
template <typename T>
T &FixedDeque<T>::front ()
{
	return mv_data[mv_firstIdx];
}


/// Get the data at the front of the deque (constant).
/**
  @return  Constant reference to the data at the front of the deque.

  @note  Calling this on an empty deque has unexpected results.
*/
template <typename T>
const T &FixedDeque<T>::front () const
{
	return mv_data[mv_firstIdx];
}


/// Get the data at the back of the deque.
/**
  @return  Reference to the data at the back of the deque.

  @note  Calling this on an empty deque has unexpected results.
*/
template <typename T>
T &FixedDeque<T>::back ()
{
	unsigned int idx = mv_lastIdx - 1;
	return mv_data[idx >= mv_size ? mv_size - 1 : idx];
}


/// Get the data at the back of the deque (constant).
/**
  @return  Constant reference to the data at the back of the deque.

  @note  Calling this on an empty deque has unexpected results.
*/
template <typename T>
const T &FixedDeque<T>::back () const
{
	unsigned int idx = mv_lastIdx - 1;
	return mv_data[idx >= mv_size ? mv_size - 1 : idx];
}


/// Get the current size of the deque.
/**
  @return  Number of items currently in the deque.
*/
template <typename T>
unsigned int FixedDeque<T>::size () const
{
	return (mv_bEmpty
		? 0
		: (mv_lastIdx <= mv_firstIdx
			? mv_size - mv_firstIdx + mv_lastIdx
			: mv_lastIdx - mv_firstIdx));
}


/// Get the maximum size of the deque.
/**
  @return  Capacity of deque.
*/
template <typename T>
unsigned int FixedDeque<T>::capacity () const
{
	return mv_size;
}


/// Check if the deque is empty.
/**
  @return  True if the deque is empty, false if not.
*/
template <typename T>
bool FixedDeque<T>::isEmpty () const
{
	return mv_bEmpty;
}


/// Check if the deque is full.
/**
  @return  True if the deque is full, false if not.
*/
template <typename T>
bool FixedDeque<T>::isFull () const
{
	return (mv_firstIdx == mv_lastIdx && !mv_bEmpty);
}


/// Assignment operator.
/**
  @param src  Source object to copy from.

  @return  Reference to this object.
*/
template <typename T>
FixedDeque<T> &FixedDeque<T>::operator = (const FixedDeque<T> &src)
{
	// Make sure we aren't trying to copy ourself.
	if (this != &src)
	{
		// Free up old memory and copy new data.
		v_free();
		v_set(src);
	}

	return *this;
}


/// Array access.
/**
  @return  Reference to the data at the requested index.

  @note  Indeces out of range have unexpected results.
*/
template <typename T>
T &FixedDeque<T>::operator [] (unsigned int idx)
{
	// Adjust index given to an index within internal array.
	idx += mv_firstIdx;
	if (idx >= mv_size || idx < mv_firstIdx)
		idx -= mv_size;

	return mv_data[idx];
}


/// Array access (constant).
/**
  @return  Constant reference to the data at the requested index.

  @note  Indeces out of range have unexpected results.
*/
template <typename T>
const T &FixedDeque<T>::operator [] (unsigned int idx) const
{
	// Adjust index given to an index within internal array.
	idx += mv_firstIdx;
	if (idx >= mv_size || idx < mv_firstIdx)
		idx -= mv_size;

	return mv_data[idx];
}


//----------------------------------------------------------------------------
// Private Methods
//----------------------------------------------------------------------------

/// Set this object's data based on data in another object.
/**
  @param src  Source object to copy from.
*/
template <typename T>
void FixedDeque<T>::v_set (const FixedDeque<T> &src)
{
	// Copy over static information.
	mv_size = src.mv_size;
	mv_firstIdx = src.mv_firstIdx;
	mv_lastIdx = src.mv_lastIdx;
	mv_bEmpty = src.mv_bEmpty;

	// Check if there's a deque to copy and copy it.
	if (mv_size)
	{
		mv_data = new T [mv_size];
		mv_dataSecond = new T [mv_size];
		memcpy(mv_data, src.mv_data, sizeof(T) * mv_size);
		memcpy(mv_dataSecond, src.mv_dataSecond, sizeof(T) * mv_size);
	}
}

/// Free allocated memory.
template <typename T>
void FixedDeque<T>::v_free ()
{
	// Delete deque data, if any.
	if (mv_data)
	{
		delete [] mv_data;
		delete [] mv_dataSecond;
	}
}

#endif	// FIXEDDEQUE_H
