/*----------------------------------------------------------------------------
  node.cpp
  Implementation for NodeFactory.

  Copyright (C) 2003 Ted Cipicchio <ted@thereisnospork.com>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to
  deal in the Software without restriction, including without limitation the
  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  IN THE SOFTWARE.
----------------------------------------------------------------------------*/
#include "stdafx.h"
#include "node.h"

//----------------------------------------------------------------------------
// Construction/Destruction
//----------------------------------------------------------------------------

/// Constructor.
/**
  @remarks  Node factories are created with an empty pool.  Nodes have to be
            allocated after factory creation using the allocate() method.

  @see allocate()
*/
NodeFactory::NodeFactory ()
	: mv_nodePoolSize(0), mv_nodePool(0),
	  mv_nodeTopIdx(0)
{
}


/// Destructor.
NodeFactory::~NodeFactory ()
{
	// Delete nodes?
	if (mv_nodePool)
	{
		while (mv_nodePoolSize--)
			if(mv_nodePool[mv_nodePoolSize]) delete mv_nodePool[mv_nodePoolSize];
		delete [] mv_nodePool;
	}
	mv_nodePool=0;
}


//----------------------------------------------------------------------------
// Public Methods
//----------------------------------------------------------------------------

/// Allocate a pool of nodes.
/**
  Nodes need to be allocated before the node factory can be used.  The node
  factory will only give out as many nodes as it has allocated (no additional
  nodes will be allocated automatically).

  @param size  Number of nodes to allocate.

  @return  True if successful, false if not.

  @pre @c size must be at least 1.

  @note  Calling allocate() after nodes have already been allocated will cause
         the old nodes to be deleted, so make sure your application is not
         using any previously fetched nodes from a factory once allocate() is
         called on it again.
*/
bool NodeFactory::allocate (unsigned int size)
{
	// Ignore sizes of 0.
	if (!size)
		return false;

	// Delete old nodes?
	if (mv_nodePool)
	{
		while (mv_nodePoolSize--)
			if(mv_nodePool[mv_nodePoolSize]) delete mv_nodePool[mv_nodePoolSize];
		delete [] mv_nodePool;
	}

	mv_nodeTopIdx = 0;
	// Allocate new nodes.
	mv_nodePool = new Node * [size];
	if(mv_nodePool)
	{
		for (unsigned int i = 0; i < size; i++)
			mv_nodePool[i] = new Node;
		// Store new size and prepare stack index.
		mv_nodePoolSize = size;
	}
	else return false;
return true;
}
