/*----------------------------------------------------------------------------
  nodehashtable-iterator.cpp
  Implementation for NodeHashIterator class.

  Copyright (C) 2003 Ted Cipicchio <ted@thereisnospork.com>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to
  deal in the Software without restriction, including without limitation the
  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  IN THE SOFTWARE.
----------------------------------------------------------------------------*/
#include "stdafx.h"
#include "nodehashtable.h"
#include "node.h"


//----------------------------------------------------------------------------
// Construction/Destruction
//----------------------------------------------------------------------------

/// Constructor.
/**
  @param pTable  Pointer to hash table to initially attach to.
                 (default = none)

  @note  The iterator will not be on any node to begin with, regardless of
         whether or not a table is initially specified.  Use the begin() or
         find() methods to place the iterator on a node.
*/
NodeHashIterator::NodeHashIterator (NodeHashTable *pTable)
	: mv_pTable(pTable),
	  mv_index(0),
	  mv_pNode(0),
	  mv_ppHolder(0)
{
}


//----------------------------------------------------------------------------
// Public Methods
//----------------------------------------------------------------------------

/// Attach this iterator to a hash table.
/**
  This will assign a hash table for this iterator to begin monitoring.

  The iterator will not be on any node to begin with.  The iterator can either
  start at the beginning of the table using the begin() method, or start at a
  node with a specific location using the find() method.

  You can also choose to detach the iterator from any hash table by passing a
  null pointer as the parameter.

  @param pTable  Table to attach the iterator to.

  @see begin(), find()
*/
void NodeHashIterator::attach (NodeHashTable *pTable)
{
	// Store the table pointer and zero out other values.
	mv_pTable = pTable;
	mv_index = 0;
	mv_pNode = 0;
	mv_ppHolder = 0;
}


/// Remove the node at the current location from the table.
/**
  @return  A pointer to the node removed from the list, or null if no pointer
           was removed.

  @remarks  This will leave the iterator invalid, and will need to be
            revalidated again after removing (either by using begin() or
            find()).  removeValid() does the same as remove(), but sets the
            iterator to the next available value (if one exists).  It is
            somewhat slower, but the iterator will still be valid as long as
			nodes are left in the table.

  @see removeValid()
*/
Node *NodeHashIterator::remove ()
{
	// Check if we even have a node.
	if (!mv_pNode)
		return 0;

	// Remove the node from the list.
	*mv_ppHolder = mv_pNode->pNext;

	// Return removed node.
	return mv_pNode;
}


/// Remove the current node from the table, keeping the iterator valid.
/**
  This will remove the current node from the table and readjust the iterator
  to the next available node.  This way the iterator can still be used to
  iterate from where if left off.

  @return  A pointer to the node removed from the list, or null if no pointer
           was removed.

  @remarks  If you want to remove a node, but do not care if the iterator is
            no longer valid afterward, use remove().  remove() is somewhat
            faster since the iterator does not need to be readjusted.

  @see remove()
*/
Node *NodeHashIterator::removeValid ()
{
	// Check if we even have a node.
	if (!mv_pNode)
		return 0;

	// Make a copy of the current node pointer and holder (so we still have
	// them after adjusting the iterator).
	Node *pold = mv_pNode;
	Node **ppholder = mv_ppHolder;

	// Set iterator to next node.
	next();

	// Remove old node from the list.
	*ppholder = pold->pNext;

	// Return removed node.
	return pold;
}


/// Return to the beginning of the table.
/**
  @return  A pointer to the node, or null if table is empty or no table is
           attached.
*/
Node *NodeHashIterator::begin ()
{
	// Check if a table is even attached.
	if (!mv_pTable)
		return 0;

	// Find first node in the table.
	unsigned int arr_size = 1 << mv_pTable->mv_nodeHashSize;
	Node **node_hash = mv_pTable->mv_nodeHash;
	for (unsigned int i = 0; i < arr_size; i++)
	{
		// Check if there are any nodes in the current list.
		Node *pcurr = node_hash[i];
		if (!pcurr)
			continue;

		// Adjust iterator values and return new current node.
		mv_pNode = pcurr;
		mv_ppHolder = &node_hash[i];
		mv_index = i;

		return pcurr;
	}

	// Table is empty, no current node.
	mv_pNode = 0;

	return 0;
}


/// Find the node at the specified location in the table.
/**
  This finds the first available node with the specified coordinates in the
  hash table and sets the iterator to the node found.

  @param x  X-coordinate of node to find.
  @param y  Y-coordinate of node to find.

  @return  The found node, or null if no node found or no table attached.

  @note  Regardless of whether or not the node is found, any node information
         in the iterator prior to the search will be changed to whatever
         results from the search.  If no node is found, the iterator will no
         longer point to any node.
*/
Node *NodeHashIterator::find (int x, int y)
{
	// If no table, no node is to be found.
	if (!mv_pTable)
		return 0;

	// Calculate the hash value of the node requested.
	unsigned int hash_val = mv_pTable->hash(x, y);

	// Scan through nodes in the list at the table location given by the hash
	// value.
	Node **node_hash = mv_pTable->mv_nodeHash;
	Node **ppholder = &node_hash[hash_val];
	Node *pnode = *ppholder;
	while (pnode && (pnode->x != x || pnode->y != y))
	{
		ppholder = &pnode->pNext;
		pnode = *ppholder;
	}

	// Store result of search.
	mv_index = hash_val;
	mv_ppHolder = ppholder;
	mv_pNode = pnode;

	return pnode;
}


/// Proceed to the next node in the table.
/**
  @return  A pointer to the new current node, or null if the end of the table
           has been reached or there is no table attached.
*/
Node *NodeHashIterator::next ()
{
	// If we have no current node, we have no next node.
	if (!mv_pNode)
		return 0;

	// Check the next node in the list.
	if (mv_pNode->pNext)
	{
		// Use next node in the list.
		mv_ppHolder = &mv_pNode->pNext;
		mv_pNode = *mv_ppHolder;

		return mv_pNode;
	}

	// No more nodes in the current list, so we must check nodes in later
	// lists.
	unsigned int arr_size = 1 << mv_pTable->mv_nodeHashSize;
	Node **node_hash = mv_pTable->mv_nodeHash;
	for (unsigned int i = mv_index + 1; i < arr_size; i++)
	{
		// Check if there are any nodes in the current list.
		Node *pcurr = node_hash[i];
		if (!pcurr)
			continue;

		// Adjust iterator values and return new current node.
		mv_pNode = pcurr;
		mv_ppHolder = &node_hash[i];
		mv_index = i;

		return pcurr;
	}

	// No more nodes, so next node is null.
	mv_pNode = 0;

	return 0;
}
