/*----------------------------------------------------------------------------
  nodehashtable.cpp
  Implementation for NodeHashTable class.

  Copyright (C) 2003 Ted Cipicchio <ted@thereisnospork.com>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to
  deal in the Software without restriction, including without limitation the
  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  IN THE SOFTWARE.
----------------------------------------------------------------------------*/
#include "stdafx.h"
#include "nodehashtable.h"

//#include <cstring>
#include "node.h"

//----------------------------------------------------------------------------
// Construction/Destruction
//----------------------------------------------------------------------------

/// Constructor.
/**
  This will create the hash table with a default size of 1 element.  You'll
  more than likely want to create a new one with better precision using the
  create() method.

  @see create()
*/
NodeHashTable::NodeHashTable ()
	: mv_nodeHashSize(0)
{
	// Create hash table with 0 bits precision (1 element).
	mv_nodeHash = new Node * [1];
	mv_nodeHash[0] = 0;
}


/// Copy constructor.
/**
  Creates a copy of another object.

  @param src  Source object to copy from.
*/
NodeHashTable::NodeHashTable (const NodeHashTable &src)
{
	// Set this object's values.
	v_set(src);
}


/// Destructor.
NodeHashTable::~NodeHashTable ()
{
	// Free up memory.
	v_free();
}


//----------------------------------------------------------------------------
// Public Methods
//----------------------------------------------------------------------------

/// Create the hash table.
/**
  This will create a hash table using the specified bits of precision.  Any
  existing hash table will be destroyed.

  The size of the table in memory will be 2^precision elements.

  @param bits  Bits of precision to use.

  @return  Precision the hash table created.

  @pre  @c bits must be an even number less than or equal to 24.  Invalid
        values will be rounded down to the nearest valid value.
*/
unsigned int NodeHashTable::create (unsigned int bits)
{
	// Adjust bit value to valid value if necessary (even number less than or
	// equal to 24).
	bits = (bits > 24 ? 24 : (bits & 0x1 ? bits & ~0x1 : bits));

	// Clear out existing table.
	delete [] mv_nodeHash;

	// Create new table.
	const size_t one = 1; // x64 compat.
	mv_nodeHash = new Node * [one << bits];
	memset(mv_nodeHash, 0, sizeof(Node *) * (one << bits));

	// Set new table size.
	mv_nodeHashSize = bits;

	return bits;
}


/// Add a node to the table.
/**
  @param pNode  Node to add.

  @note  Null pointers are ignored.
*/
bool NodeHashTable::add (Node *pNode)
{
	// Check for null pointer.
	if (!pNode)
		return false;

	// Get the hash value for this node.
	unsigned int hash_val = hash(pNode->x, pNode->y);

	// GF: Don't add if already inside!
	for(Node *pcurr = mv_nodeHash[hash_val]; pcurr; pcurr = pcurr->pNext)
	{
		if(pcurr == pNode) return false;
	}

	// Insert this node into the list at the location in the table given by
	// the node's hash value.
	Node *&rpdest = mv_nodeHash[hash_val];
	pNode->pNext = rpdest;
	rpdest = pNode;
return true;
}


/// Get the number of nodes in the hash table.
/**
  @return  Number of nodes in the hash table.
*/
unsigned int NodeHashTable::size () const
{
	// Start count at 0...
	unsigned int count = 0;

	// Loop through each array element.
	unsigned int arr_size = 1U << mv_nodeHashSize;
	while (arr_size--)
	{
		// Start at the beginning of the list.
		Node *pcurr = mv_nodeHash[arr_size];
		while (pcurr)
		{
			// Node here, so increment count.
			count++;

			// Continue onto next node.
			pcurr = pcurr->pNext;
		}
	}

	return count;
}


/// Empty the hash table.
/**
  This will empty out all entries in the hash table.  No memory is deleted or
  nodes released when doing so.

  If you want to empty out the table and release all nodes at the same time,
  use the releaseAll() method.

  @see releaseAll()
*/
void NodeHashTable::empty ()
{
	const size_t one = 1; // x64 compat.
	memset(mv_nodeHash, 0, sizeof(Node *) * (one << mv_nodeHashSize));
}


/// Release all nodes in the hash table.
/**
  This will empty out all entries in the hash table and release all nodes back
  into the specified node factory.  No memory is deleted when doing so.

  If you want to empty out the table without releasing nodes, use the empty()
  method.

  @param factory  Factory to release nodes back into.

  @see empty()
*/
void NodeHashTable::releaseAll (NodeFactory &factory)
{
	// Cycle through each entry in the table.
	unsigned int idx = 1 << mv_nodeHashSize;
	while (idx--)
	{
		// Release all nodes in the current list.
		Node *pnode = mv_nodeHash[idx];
		while (pnode)
		{
			Node *pnext = pnode->pNext;
			factory.release(pnode);
			pnode = pnext;
		}
	}

	// Clear out the array.
	const size_t one = 1; // x64 compat.
	memset(mv_nodeHash, 0, sizeof(Node *) * (one << mv_nodeHashSize));
}


/// Node hash function.
/**
  This calculates the hash value of a node with the specified coordinates.

  This is mainly used internally, but is available publicly in case one wants
  to play with it.

  @param x  X-coordinate of the node.
  @param y  Y-coordinate of the node.

  @return  Hash code for the node.

  @remarks  The hash table makes use of the coordinates of the node to
            calculate the node's hash value.  The lower-order bits of each
            coordinate are combined to get the resulting hash code for the
            node.  The number of bits used depends on the precision of the
            table created (i.e. a 16-bit table will use the lower 8-bits of
            each coordinate).
*/
unsigned int NodeHashTable::hash (int x, int y) const
{
	// Generate and return code.
	unsigned int half_size = mv_nodeHashSize >> 1;
	int mask = (1 << half_size) - 1;
	return (x & mask) | ((y & mask) << half_size);
}


/// Assignment operator.
/**
  Copies data from one object into this one.

  @param src  Source object to copy from.

  @return  Reference to this object.
*/
NodeHashTable &NodeHashTable::operator = (const NodeHashTable &src)
{
	// Make sure we aren't trying to copy ourself.
	if (this != &src)
	{
		// Free old data and copy.
		v_free();
		v_set(src);
	}

	return *this;
}


//----------------------------------------------------------------------------
// Private Methods
//----------------------------------------------------------------------------

/// Set object data based on data from another object.
/**
  @param src  Source object to copy from.
*/
void NodeHashTable::v_set (const NodeHashTable &src)
{
	// Get size of table.
	mv_nodeHashSize = src.mv_nodeHashSize;

	// Allocate memory for table and copy.
	const size_t one = 1; // x64 compat.
	mv_nodeHash = new Node * [one << mv_nodeHashSize];
	memcpy(mv_nodeHash, src.mv_nodeHash, one << mv_nodeHashSize);
}


/// Free allocated memory.
void NodeHashTable::v_free ()
{
	// Delete hash table.
	if(mv_nodeHash)delete [] mv_nodeHash;
}
