#ifndef NODEHASHTABLE_H
#define NODEHASHTABLE_H
/**
  @class NodeHashTable
  @brief Hash table for path planner nodes.

  This is a hash table specifically designed for path-planner nodes.  Hash
  codes are generated based on the node location.

  Copyright (C) 2003 Ted Cipicchio <ted@thereisnospork.com>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to
  deal in the Software without restriction, including without limitation the
  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  IN THE SOFTWARE.

  @date  Created: Thu Mar 20 2003
*/
/**
  @class NodeHashIterator
  @brief Iterator for path planner node hash table.

  Copyright (C) 2003 Ted Cipicchio <ted@thereisnospork.com>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to
  deal in the Software without restriction, including without limitation the
  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  IN THE SOFTWARE.

  @date  Created: Thu Mar 20 2003
*/

struct Node;
class NodeFactory;

//----------------------------------------------------------------------------
class NodeHashTable
{
	friend class NodeHashIterator;

public:
	//------------------------------------------------------------------------
	/** @name Construction/Destruction */
	//@{

	/// Constructor.
	NodeHashTable ();
	/// Copy constructor.
	NodeHashTable (const NodeHashTable &src);
	/// Destructor.
	virtual ~NodeHashTable ();

	//@}


	//------------------------------------------------------------------------
	/** @name General Operations */
	//@{

	/// Create the hash table.
	unsigned int create (unsigned int bits);

	/// Add a node to the table.
	bool add (Node *pNode);

	/// Get the number of nodes in the hash table.
	unsigned int size () const;

	/// Empty the hash table.
	void empty ();
	/// Release all nodes in the hash table.
	void releaseAll (NodeFactory &factory);

	/// Node hash function.
	unsigned int hash (int x, int y) const;

	//@}


	//------------------------------------------------------------------------
	/** @name Overloaded Operators */
	//@{

	/// Assignment operator.
	NodeHashTable &operator = (const NodeHashTable &src);

	//@}

private:
	/// Hash table (array of node linked lists).
	Node **mv_nodeHash;
	/// Hash table size.
	unsigned int mv_nodeHashSize;


	/// Set object data based on data from another object.
	void v_set (const NodeHashTable &src);
	/// Free allocated memory.
	void v_free ();
};


//----------------------------------------------------------------------------
class NodeHashIterator
{
public:
	/// Constructor.
	NodeHashIterator (NodeHashTable *pTable = 0);

	/// Attach this iterator to a hash table.
	void attach (NodeHashTable *pTable);

	/// Get the node at the current location.
	const Node *get () const;

	/// Remove the node at the current location from the table.
	Node *remove ();
	/// Remove the current node from the table, keeping the iterator valid.
	Node *removeValid ();

	/// Return to the beginning of the table.
	Node *begin ();

	/// Find the node at the specified location in the table.
	Node *find (int x, int y);

	/// Proceed to the next node in the table.
	Node *next ();

private:
	/// Pointer to hash table this is iterating.
	NodeHashTable *mv_pTable;
	/// Hash table index.
	unsigned int mv_index;
	/// Pointer to current node.
	Node *mv_pNode;
	/// Pointer to where the current node is held.
	Node **mv_ppHolder;
};


/// Get the node at the current location.
/**
  @return  Pointer to the node at the current location, or null if the
           iterator has reached the end of the table.
*/
inline const Node *NodeHashIterator::get () const
{
	return mv_pNode;
}

#endif	// NODEHASHTABLE_H
