/*----------------------------------------------------------------------------
  planner.cpp
  Implementation for Planner class.

  Copyright (C) 2003 Ted Cipicchio <ted@thereisnospork.com>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to
  deal in the Software without restriction, including without limitation the
  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  IN THE SOFTWARE.
----------------------------------------------------------------------------*/
#include "stdafx.h"
#include "planner.h"


// Maximum value for color values (3 * 255).
// #define COLORVAL_MAX 765


//----------------------------------------------------------------------------
// Construction/Destruction
//----------------------------------------------------------------------------

/// Constructor.
Planner::Planner ()
	: mt_bDone(true), mt_map(0), mt_numSlices(0), mt_runningTime(0),
	  mt_pGoal(0)
{
	// Whoosh...
	memset(&mt_description, 0, sizeof(Description));
}


/// Destructor.
Planner::~Planner ()
{
	// Make sure path info is cleaned up.
	cleanup();

	// Clear out old data.
	if (mt_map)
		delete [] mt_map;
}


//----------------------------------------------------------------------------
// Construction/Destruction
//----------------------------------------------------------------------------

/// Create a map to use for plotting.
/**
  @param image    BMP image to use as map.
  @param descrip  Description of how to interpret the map and how to set up
                  the algorithm.

  @return  True if successful, false if failed (no map set up).
*/
bool Planner::createMap(const Description &descrip)
{
	// Make sure the start and end locations are valid.
	if (descrip.startX < 0 || descrip.startX >= (int)descrip.width ||
		descrip.startY < 0 || descrip.startY >= (int)descrip.height ||
		descrip.endX < 0 || descrip.endX >= (int)descrip.width ||
		descrip.endY < 0 || descrip.endY >= (int)descrip.height)
		return false;

	// Make sure start and end locations are not the same.
	if (descrip.startX == descrip.endX && descrip.startY == descrip.endY)
		return false;

	// Make sure the threshold for walls is reasonable.
	// if (descrip.wallThreshold < 0 || descrip.wallThreshold >= COLORVAL_MAX - 1) return false;

	// Make sure multiplier for ground weighting is not negative.
	// if (descrip.weightMultiplier < 0.0f) return false;

	// Unload the old map.
	unloadMap();

	// Create the array for our map.
	mt_map = new float [(descrip.width) * (descrip.height)];

	// Read in values from double-array.
	long idx = 0;
	float val, maxval = -1.0;
	for (unsigned long y = 0; y < descrip.height; y++)
	{
		for (unsigned long x = 0; x < descrip.width; x++)
		{
			val = (float) *(descrip.pMapData + x + descrip.width * y);
			// Negative value? -> Becomes impermiable
			if (val <= 0.0f)
			{
				mt_map[idx++] = -1.0f;
				continue;
			}
			if(x==0 && y==0) maxval = val; // first value
			if (val>maxval) maxval=val;
			mt_map[idx++] = val;
		}
	}

	// Normalize to values from 1.0 to 2.0 OR -1.0 for solid
	if (maxval>0)
	{
		for (unsigned long y=0; y<descrip.width * descrip.height; y++)
			if(mt_map[y]>0)
			{
				mt_map[y] = mt_map[y]/maxval + 1.0f;
			}
	}
	else return false;

	// Allocate nodes in factory.
	mt_nodeFactory.allocate(descrip.factorySize
		? descrip.factorySize
		: descrip.width * descrip.height);

	// Allocate deque memory.
	mt_open.allocate(descrip.openDequeSize
		? descrip.openDequeSize
		: descrip.width * descrip.height);

	// Initialize hash table for closed nodes.
	mt_closed.create(descrip.hashPrecision);

	// No goal node yet...
	mt_pGoal = 0;

	// Store description...
	mt_description = descrip;
	mt_description.width = descrip.width;
	mt_description.height = descrip.height;

	// Ensure heuristic weight given is within the range [0, 1].
	if (mt_description.heuristicWeight < 0.0f)
		mt_description.heuristicWeight = 0.0f;
	else if (mt_description.heuristicWeight > 1.0f)
		mt_description.heuristicWeight = 1.0f;

	// Reset number of slices run to 0.
	mt_numSlices = 0;

	// No time run so far.
	mt_runningTime = 0;

	// Say we're "done" so path-planning isn't run before calling
	// initialize().
	mt_bDone = true;

	return true;
}


/// Unload any currently loaded map.
void Planner::unloadMap ()
{
	// Check if a map even exists.
	if (!mt_map)
		return;

	// Delete the map.
	delete [] mt_map;
	mt_map = 0;

	// Zero out number of slices.
	mt_numSlices = 0;

	// No goal.
	mt_pGoal = 0;

	// Clear out description.
	memset(&mt_description, 0, sizeof(Description));
}


/// Initialize the path-finding algorithm.
/**
  This sets up the state of varables needed to begin path finding.

  The default implementation of this function unsets the "done" state and sets
  the current number of slices run to 0.  Overrides of this function should be
  sure to call the base class version.

  @note  If a path algorithm is to be executed multiple times on the same
         data, cleanup() should be called before calling initialize().

  @see run(), cleanup()
*/
void Planner::initialize ()
{
	// Not done.
	mt_bDone = false;

	// No slices yet...
	mt_numSlices = 0;

	// No goal (how many times am I gonna set this to 0?).
	mt_pGoal = 0;
}


/// Clean up nodes in use.
/**
  This empties out the open and closed lists and makes sure the nodes are
  released back into the factory.  This is necessary to ensure nodes are
  properly deleted upon destruction.

  The default implementation of this function removes and frees all the nodes
  from the open and closed lists, and clears out the path list.  The path is
  assumed to represent nodes within the closed list, and as such its nodes are
  not released.  Overrides of this function should be sure to call the base
  class version.

  @see initialize(), run()
*/
void Planner::cleanup ()
{
	// Release open nodes.
	for (unsigned int i = 0; i < mt_open.size(); i++)
	{
		Node* pN = mt_open[i];
		mt_nodeFactory.release(pN);
	}
	mt_open.clear();

	// Release closed nodes.
	mt_closed.releaseAll(mt_nodeFactory);

	// Clear path list.
	mt_pGoal = 0;

	// Set done state to keep from accidentally running again.
	mt_bDone = true;
}


/// Get path cost.
/**
  This gets the cost of the final path, if one exists.  The cost is pulled
  from the cost stored in the last node of the path.

  @return  The final path cost, or -1 if no path.
*/
float Planner::getPathCost () const
{
	// Check for a path.
	if (!mt_pGoal)
		return -1.0f;

	return mt_pGoal->cost;
}


/// Get the size of the path list.
/**
  @return  Number of nodes currently in the path list.
*/
unsigned int Planner::getPathSize () const
{
	// Count the size.
	unsigned int path_size = 0;
	Node *pnode = mt_pGoal;
	while (pnode)
	{
		path_size++;
		pnode = pnode->pParent;
	}

	return path_size;
}


/// Get the size of the open list.
/**
  @return  Number of nodes currently in the open list.
*/
unsigned int Planner::getOpenSize () const
{
	return mt_open.size();
}


/// Get the size of the closed list.
/**
  @return  Number of nodes currently in the closed list.
*/
unsigned int Planner::getClosedSize () const
{
	return mt_closed.size();
}

