#ifndef TIMER_H
#define TIMER_H
/**
  @class Timer
  @brief Cross-platform timer class.

  This provides a general-purpose timer, with time given in milliseconds.

  Copyright (C) 2003 Ted Cipicchio <ted@thereisnospork.com>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to
  deal in the Software without restriction, including without limitation the
  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  IN THE SOFTWARE.

  @date  Created: Sun Mar 23 2003

  @note  If this is compiled in a project using SDL, the macro @c WANT_SDL
         should be defined so SDL's timer functions are used.  Otherwise, the
         timer will attempt to use the performance counter functions provided
         by the Windows API.

  @note  From an OO perspective, I could have used a generic interface for the
         timer and left the cross-platform issues up to the implementations,
         but I wanted the class to run as fast as possible so as not to impact
         the application's performance (primarily with the get() method).
*/


#ifdef WANT_SDL
#include <SDL/SDL.h>
#else
#include "stdafx.h"
#endif


//----------------------------------------------------------------------------
class Timer
{
public:
	//------------------------------------------------------------------------
	/** @name Construction/Destruction */
	//@{

	/// Constructor.
	Timer ();

	//@}


	//------------------------------------------------------------------------
	/** @name General Operations */
	//@{

	/// Start the timer.
	void start ();
	/// Stop the timer.
	void stop ();

	/// Check if the timer is running.
	bool isRunning () const;

	/// Get the current timer reading.
	unsigned int get () const;

	/// Reset the timer.
	bool reset ();

	//@}

private:
	/// Time accumulated up to start time.
	unsigned int mv_accumulatedTime;
	/// True if timer is running, false if not.
	bool mv_bRunning;

#ifdef WANT_SDL
	/// Timer start time.
	unsigned int mv_startTime;
#else
	/// Timer start time.
	unsigned int mv_startTime;
	/// Scale factor to convert performance counter reading to milliseconds.
	double mv_counterScale;
#endif
};


/// Constructor.
/**
  The timer is initialized with a time of 0 upon creation.
*/
inline Timer::Timer ()
	: mv_accumulatedTime(0), mv_bRunning(false)
{
#ifndef WANT_SDL
	// Create scale factor from performance counter frequency to get msec from
	// counter queries.
	unsigned int perf_freq;
	QueryPerformanceFrequency(&perf_freq);
	mv_counterScale = 1000.0 / double(perf_freq.QuadPart);
#endif
}


/// Start the timer.
/**
  This starts the timer running from its current time.  Timers can be
  restarted after they have been stopped and continue from the stop time.
  Starting the timer while it's running will have no effect.
*/
inline void Timer::start ()
{
	// Ignore if timer is running.
	if (mv_bRunning)
		return;

	// Get timer time and set running flag.
#ifdef WANT_SDL
	mv_startTime = SDL_GetTicks();
#else
	QueryPerformanceCounter(&mv_startTime);
#endif
	mv_bRunning = true;
}


/// Stop the timer.
/**
  This stops the timer at its current time.  The timer can then either be
  reset or started again at a later time.
*/
inline void Timer::stop ()
{
	// Ignore if timer is not running.
	if (!mv_bRunning)
		return;

	// Add time since start time to accumulated time.
#ifdef WANT_SDL
	mv_accumulatedTime += SDL_GetTicks() - mv_startTime;
#else
	unsigned int curr_time;
	QueryPerformanceCounter(&curr_time);
	mv_accumulatedTime +=
		(unsigned int)(double(curr_time.QuadPart - mv_startTime.QuadPart)
			* mv_counterScale);
#endif

	// Unset running flag.
	mv_bRunning = false;
}


/// Check if the timer is running.
/**
  @return  True if the timer is currently running, false if not.
*/
inline bool Timer::isRunning () const
{
	return mv_bRunning;
}


/// Get the current timer reading.
/**
  This will return the time currently accumulated by the timer, in
  milliseconds.  Timer readings can be checked both while the timer is running
  and while it is stopped.

  @return  Current timer reading, in milliseconds.
*/
inline unsigned int Timer::get () const
{
	// If timer isn't running, just return accumulated time.
	if (!mv_bRunning)
		return mv_accumulatedTime;

	// Timer time is accumulated time plus time since timer start.
#ifdef WANT_SDL
	return mv_accumulatedTime + (SDL_GetTicks() - mv_startTime);
#else
	unsigned int curr_time;
	QueryPerformanceCounter(&curr_time);
	return mv_accumulatedTime +
		(unsigned int)(double(curr_time.QuadPart - mv_startTime.QuadPart)
			* mv_counterScale);
#endif
}


/// Reset the timer.
/**
  This will reset the timer to 0.  The timer will only be reset if it is not
  currently running.

  @return  True if the timer was reset, false if not (timer was running).
*/
inline bool Timer::reset ()
{
	// Make sure timer isn't running.
	if (mv_bRunning)
		return false;

	// Reset time to zero.
	mv_accumulatedTime = 0;

	return true;
}

#endif	// TIMER_H
