/*----------------------------------------------------------------------------
  breadthplanner.cpp
  Implementation for BreadthPlanner class.

  Copyright (C) 2003 Ted Cipicchio <ted@thereisnospork.com>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to
  deal in the Software without restriction, including without limitation the
  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  IN THE SOFTWARE.
----------------------------------------------------------------------------*/
#include "breadthplanner.h"


// Check the specified node against the open and closed lists and add to open.
#define CHECK_ADD(X, Y, MAP_IDX) \
	do                                                           \
	{                                                            \
		/* Make sure we're not off the map. */                   \
		if ((X) < 0 || (X) >= width || (Y) < 0 || (Y) >= height) \
			break;                                               \
		                                                         \
		/* Check for wall. */                                    \
		if (map[MAP_IDX] == PLANNER_WALL)                        \
			break;                                               \
		                                                         \
		/* Check closed list. */                                 \
		if (closed_iter.find(X, Y))                              \
			break;                                               \
		                                                         \
		/* Check open list. */                                   \
		unsigned int i = open.size();                            \
		bool bstop = false;                                      \
		while (i--)                                              \
		{                                                        \
			Node &rcurr = *open[i];                              \
			if ((X) == rcurr.x && (Y) == rcurr.y)                \
			{                                                    \
				bstop = true;                                    \
				break;                                           \
			}                                                    \
		}                                                        \
		if (bstop)                                               \
			break;                                               \
		                                                         \
		/* Not in either list, so it should be added. */         \
		Node *pchild = factory.fetch();                          \
		Node &rchild = *pchild;                                  \
		rchild.pParent = pnode;                                  \
		rchild.x = X;                                            \
		rchild.y = Y;                                            \
		rchild.pNext = 0;                                        \
		open.pushBack(pchild);                                   \
	} while (false);


//----------------------------------------------------------------------------
// Public Methods
//----------------------------------------------------------------------------

/// Initialize the path-finding algorithm.
/**
  @copydoc Planner::initialize()
*/
void BreadthPlanner::initialize ()
{
	// Base class initialize()...
	Planner::initialize();

	// Get node for starting point and put into queue.
	Node *pnode = mt_nodeFactory.fetch();
	pnode->pParent = 0;
	pnode->x = mt_description.startX;
	pnode->y = mt_description.startY;
	pnode->pNext = 0;
	mt_open.pushBack(pnode);
}


/// Run a single time slice of the path-finding algorithm.
/*
  @copydoc Planner::run()
*/
bool BreadthPlanner::run ()
{
	// Increment run count.
	mt_numSlices++;

	// Get some useful variables.
	int end_x = mt_description.endX;
	int end_y = mt_description.endY;
	unsigned int width  = mt_description.width;
	unsigned int height = mt_description.height;
	float *map = mt_map;
	FixedDeque<Node *> &open = mt_open;
	NodeHashTable &closed = mt_closed;
	NodeHashIterator closed_iter(&closed);
	NodeFactory &factory = mt_nodeFactory;
//	unsigned int slice_time = mt_description.sliceTime;

	// Create a timer and start it.
//	Timer timer;
//	timer.start();

	// Cycle through open nodes.
	while (!open.isEmpty())
	{
		// Get the next node to examine.
		Node *pnode = open.front();
		open.popFront();
		Node &rnode = *pnode;
		int x = rnode.x;
		int y = rnode.y;

		// Check if the node is the end node.
		if (x == end_x && y == end_y)
		{
			closed.add(pnode);
			mt_pGoal = pnode;
//			timer.stop();
//			mt_runningTime += timer.get();

			mt_bDone = true;

			return true;
		}

		unsigned int map_idx = x + y * width;

		// Add this node's children to the open list.
		int child_x = x - 1;
		int child_y = y - 1;
		CHECK_ADD(child_x, child_y, map_idx - width - 1)
		child_x++;
		CHECK_ADD(child_x, child_y, map_idx - width)
		child_x++;
		CHECK_ADD(child_x, child_y, map_idx - width + 1)
		child_y++;
		CHECK_ADD(child_x, child_y, map_idx + 1)
		child_x -= 2;
		CHECK_ADD(child_x, child_y, map_idx - 1)
		child_y++;
		CHECK_ADD(child_x, child_y, map_idx + width - 1)
		child_x++;
		CHECK_ADD(child_x, child_y, map_idx + width)
		child_x++;
		CHECK_ADD(child_x, child_y, map_idx + width + 1)

		// Add it to the closed list.
		closed.add(pnode);

		// Check if we've run out of time for this slice.
//		if (slice_time && timer.get() > slice_time)	break;
	}

//	timer.stop();
//	mt_runningTime += timer.get();

	mt_bDone = open.isEmpty();

	return mt_bDone;
}
