#ifndef VECTOR_MATH_H_INCLUDED
#define VECTOR_MATH_H_INCLUDED

#include <math.h>
#include <memory.h>

#ifndef SQUARE
#define SQUARE(a) ((a)*(a))
#endif

// Fast math replacements:
/*
 * Code for approximation of cos, sin, tan and inv sin, etc.
 * Surprisingly accurate and very usable.
 *
 * Domain:
 * Sin/Cos    [0, pi/2]
 * Tan        [0,pi/4]
 * InvSin/Cos [0, 1]
 * InvTan     [-1, 1]
 */
#define  SGD_PI_2  1.57079632679489661923f

// --------------------------------------------------------------- //
// use sqrtf for accuracy. This is about 5% off the real value
// --------------------------------------------------------------- //
inline float qsqrt(const float f)
{
    unsigned long l;
    l=*((unsigned long*)(&f));
    l-=0x3F800000l;
    l>>=(1);
    l+=0x3F800000l;
    *((unsigned long*)(&f))=l;
    return f;
}

inline float qsin(const float val)
{
    float fASqr = val*val;
    float fResult = -2.39e-08f;
    fResult *= fASqr;
    fResult += 2.7526e-06f;
    fResult *= fASqr;
    fResult -= 1.98409e-04f;
    fResult *= fASqr;
    fResult += 8.3333315e-03f;
    fResult *= fASqr;
    fResult -= 1.666666664e-01f;
    fResult *= fASqr;
    fResult += 1.0f;
    fResult *= val;

    return fResult;
}

inline float qcos(const float val)
{
    float fASqr = val*val;
    float fResult = -2.605e-07f;
    fResult *= fASqr;
    fResult += 2.47609e-05f;
    fResult *= fASqr;
    fResult -= 1.3888397e-03f;
    fResult *= fASqr;
    fResult += 4.16666418e-02f;
    fResult *= fASqr;
    fResult -= 4.999999963e-01f;
    fResult *= fASqr;
    fResult += 1.0f;

    return fResult;  
}

inline float qtan(const float val)
{
    float fASqr = val*val;
    float fResult = 9.5168091e-03f;
    fResult *= fASqr;
    fResult += 2.900525e-03f;
    fResult *= fASqr;
    fResult += 2.45650893e-02f;
    fResult *= fASqr;
    fResult += 5.33740603e-02f;
    fResult *= fASqr;
    fResult += 1.333923995e-01f;
    fResult *= fASqr;
    fResult += 3.333314036e-01f;
    fResult *= fASqr;
    fResult += 1.0f;
    fResult *= val;

    return fResult;
}

inline float qasin(float val)
{
    float fRoot = qsqrt(1.0f-val);
    float fResult = -0.0187293f;
    fResult *= val;
    fResult += 0.0742610f;
    fResult *= val;
    fResult -= 0.2121144f;
    fResult *= val;
    fResult += 1.5707288f;
    fResult = SGD_PI_2 - fRoot*fResult;

    return fResult;
}

inline float qacos(float val)
{
    float fRoot = qsqrt(1.0f-val);
    float fResult = -0.0187293f;
    fResult *= val;
    fResult += 0.0742610f;
    fResult *= val;
    fResult -= 0.2121144f;
    fResult *= val;
    fResult += 1.5707288f;
    fResult *= fRoot;

    return fResult;
}

inline float qatan(float val)
{
    float fVSqr = val*val;
    float fResult = 0.0028662257f;
    fResult *= fVSqr;
    fResult -= 0.0161657367f;
    fResult *= fVSqr;
    fResult += 0.0429096138f;
    fResult *= fVSqr;
    fResult -= 0.0752896400f;
    fResult *= fVSqr;
    fResult += 0.1065626393f;
    fResult *= fVSqr;
    fResult -= 0.1420889944f;
    fResult *= fVSqr;
    fResult += 0.1999355085f;
    fResult *= fVSqr;
    fResult -= 0.3333314528f;
    fResult *= fVSqr;
    fResult += 1.0f;
    fResult *= val;

    return fResult;
}


// -------------------------------------------------------------- //
// gSetVertex: Assign
// -------------------------------------------------------------- //
inline float* gSetVertex(float* vertex, float x=0.0f, float y=0.0f, float z=0.0f)
{
	vertex[0]=x; vertex[1]=y; vertex[2]=z;
	return vertex;
}


// -------------------------------------------------------------- //
// gGetLength: Get Squared Length of Vetrex
// -------------------------------------------------------------- //
inline float gGetSQLength(const float* vertex)
{
	// calculate the length of the normal
	return	(vertex[0])*(vertex[0]) + 
			(vertex[1])*(vertex[1]) + 
			(vertex[2])*(vertex[2]);
}

// -------------------------------------------------------------- //
// gGetLength: Get Length of Vetrex
// -------------------------------------------------------------- //
inline float gGetLength(const float* vertex)
{
	// calculate the length of the normal
	float sql = gGetSQLength(vertex);
	if(sql>1E-7)
		return (float)sqrt(sql);
return 0;
}

// -------------------------------------------------------------- //
// gDot: Dot product
// -------------------------------------------------------------- //
inline float gDot(const float* v1, const float* v2)
{
	return v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2];
}

// -------------------------------------------------------------- //
// gCross: Cross product
// -------------------------------------------------------------- //
inline void gCross(const float* v1, const float* v2, float* res)
{
	res[0] = v1[1]*v2[2] - v1[2]*v2[1];
	res[1] = v1[2]*v2[0] - v1[0]*v2[2];
	res[2] = v1[0]*v2[1] - v1[1]*v2[0];
}

// -------------------------------------------------------------- //
// gCopyVertex: Copy one vector on another
// -------------------------------------------------------------- //
inline void gCopyVertex(float* pdest, const float* psrc)
{
	*(pdest++)=*(psrc++);
	*(pdest++)=*(psrc++);
	*(pdest  )=*(psrc  );
}

// -------------------------------------------------------------- //
// gCopyMatrix: Copy 4x4 matrix
// -------------------------------------------------------------- //
inline void gCopyMatrix(float* pdest, const float* psrc)
{
	*(pdest++) = *(psrc++);	*(pdest++) = *(psrc++);	*(pdest++) = *(psrc++);	*(pdest++) = *(psrc++);
	*(pdest++) = *(psrc++);	*(pdest++) = *(psrc++);	*(pdest++) = *(psrc++);	*(pdest++) = *(psrc++);
	*(pdest++) = *(psrc++);	*(pdest++) = *(psrc++);	*(pdest++) = *(psrc++);	*(pdest++) = *(psrc++);
	*(pdest++) = *(psrc++);	*(pdest++) = *(psrc++);	*(pdest++) = *(psrc++);	*(pdest++) = *(psrc++);
}

// -------------------------------------------------------------- //
// gAddVertex: Add one vertex to another
// -------------------------------------------------------------- //
inline void gAddVertex(float* v1, const float* v2)
{
	*(v1++)+=*(v2++);
	*(v1++)+=*(v2++);
	*(v1  )+=*(v2  );
}
// -------------------------------------------------------------- //
// gAddVertex: Subtract one vertex from another
// -------------------------------------------------------------- //
inline void gSubVertex(float* base, const float* vsub)
{
	*(base++)-=*(vsub++);
	*(base++)-=*(vsub++);
	*(base  )-=*(vsub  );
}

// -------------------------------------------------------------- //
// gDiffVertex: res = v1-v2
// -------------------------------------------------------------- //
inline void gDiffVertex(float* res, const float* v1, const float* v2)
{
	res[0] = v1[0]- v2[0];
	res[1] = v1[1]- v2[1];
	res[2] = v1[2]- v2[2];
}

// -------------------------------------------------------------- //
// gAddVertex: Add one vertex to another
// -------------------------------------------------------------- //
inline void gMultVertex(float* v1, const float scale)
{
	*(v1++)*=scale;
	*(v1++)*=scale;
	*(v1  )*=scale;
}

// -------------------------------------------------------------- //
// Reverse a vertex
// -------------------------------------------------------------- //
inline void gFlipVertex(float* v)
	{v[0]=-v[0]; v[1]=-v[1]; v[2]=-v[2];}

// -------------------------------------------------------------- //
// gUnify: Make unit vector
// -------------------------------------------------------------- //
inline void gUnify(float* v1)
{
float	len = gGetLength(v1);
	if (len!=0.0f)
	{
		*(v1++)/=len;
		*(v1++)/=len;
		*(v1  )/=len;
	}
}


// -------------------------------------------------------------- //
// gInvertMatrix4x4: Invert a 4x4 matrix
// only for symetrical case
// -------------------------------------------------------------- //
inline void gInvertMatrix4x4(float m[16])
{
static float temp[16];
	temp[0]=m[0];
	temp[1]=m[4];
	temp[2]=m[8];
	temp[3]=0.0f;
	temp[4]=m[1];
	temp[5]=m[5];
	temp[6]=m[9];
	temp[7]=0.0f;
	temp[8]=m[2];
	temp[9]=m[6];
	temp[10]=m[10];
	temp[11]=0.0f;
	temp[12]=-(m[12]*m[0])-(m[13]*m[1])-(m[14]*m[2]);
	temp[13]=-(m[12]*m[4])-(m[13]*m[5])-(m[14]*m[6]);
	temp[14]=-(m[12]*m[8])-(m[13]*m[9])-(m[14]*m[10]);
	temp[15]=1.0f;
	const float*pR=&temp[0];
	float* pW=&m[0];
	*pW++ = *pR++; *pW++ = *pR++; *pW++ = *pR++; *pW++ = *pR++; 
	*pW++ = *pR++; *pW++ = *pR++; *pW++ = *pR++; *pW++ = *pR++; 
	*pW++ = *pR++; *pW++ = *pR++; *pW++ = *pR++; *pW++ = *pR++; 
	*pW++ = *pR++; *pW++ = *pR++; *pW++ = *pR++; *pW++ = *pR++; 
}

inline void gIdentity(float m[16]) 
{
	 m[0] = 1; m[4] = 0; m[ 8] = 0; m[12] = 0;
	 m[1] = 0; m[5] = 1; m[ 9] = 0; m[13] = 0;
	 m[2] = 0; m[6] = 0; m[10] = 1; m[14] = 0;
	 m[3] = 0; m[7] = 0; m[11] = 0; m[15] = 1; 
}

inline void gInvertMatrix4x4_full(float src[16])
{
	float inverse[16];
	float t; int i, j, k, swap; float tmp[4][4];
	gIdentity(inverse);
	for (i = 0; i < 4; i++)
	{
		 for (j = 0; j < 4; j++)
		{
			 tmp[i][j] = src[i*4+j]; 
		} 
	}
	for (i = 0; i < 4; i++)
	{
		// look for largest element in column.
		swap = i;
		for (j = i + 1; j < 4; j++) 
		{
			if (fabs(tmp[j][i]) > fabs(tmp[i][i])) 
			{
				 swap = j; 
			} 
		}
		if (swap != i) 
		{
			 // swap rows.
			for (k = 0; k < 4; k++) 
			{
				 t = tmp[i][k]; tmp[i][k] = tmp[swap][k]; tmp[swap][k] = t; t = inverse[i*4+k]; inverse[i*4+k] = inverse[swap*4+k]; inverse[swap*4+k] = t; 
			} 
		}
		if (tmp[i][i] == 0) 
		{
			 // no non-zero pivot. the matrix is singular, which shouldn't happen. This means the user gave us a bad matrix.
			 return;
		}
		t = tmp[i][i];
		for (k = 0; k < 4; k++) 
		{
			 tmp[i][k] /= t; inverse[i*4+k] /= t; 
		}
		for (j = 0; j < 4; j++) 
		{
			if (j != i) 
			{
				t = tmp[j][i];
				for (k = 0; k < 4; k++) 
				{
					 tmp[j][k] -= tmp[i][k]*t; inverse[j*4+k] -= inverse[i*4+k]*t; 
				} 
			} 
		} 
	}
	memcpy(src, inverse, 16*sizeof(float));
}

// -------------------------------------------------------------- //
// gComputeNormal:
// -------------------------------------------------------------- //
inline void gComputeNormal(const float LastVectors[3][3], float* pNormal)
{
float	a[3];// = {0.0f, 0.0f, 0.0f};
float	b[3];// = {0.0f, 0.0f, 0.0f};
float	result[3];// = {0.0f, 0.0f, 0.0f};

	b[0]=(LastVectors[1][0] - LastVectors[0][0]);
	b[1]=(LastVectors[1][1] - LastVectors[0][1]);
	b[2]=(LastVectors[1][2] - LastVectors[0][2]);

	a[0]=-(LastVectors[1][0] - LastVectors[2][0]);
	a[1]=-(LastVectors[1][1] - LastVectors[2][1]);
	a[2]=-(LastVectors[1][2] - LastVectors[2][2]);

	// a x b
	result[0]= ((a[1]*b[2]) - (b[1]*a[2]));
	result[1]= ((a[2]*b[0]) - (b[2]*a[0]));
	result[2]= ((a[0]*b[1]) - (b[0]*a[1]));

	gCopyVertex(pNormal, result);
	return;

/*
float	length;
	// calculate the length of the normal
	length=gGetLength(result);

	// normalize and specify the normal
	if (length!=0.0f)
	{
		length=1.0f;
		pNormal[0]=result[0]/length;
		pNormal[1]=result[1]/length;
		pNormal[2]=result[2]/length;
	}
	else
	{
		memset(pNormal, 0, sizeof(float)*3);
	}
*/
}

inline void gComputeTangentBasis(const float LastVectors[3][3], const float LastTextures[3][2], float *tangent, float *binormal)
{
float cp[3];
float e0[3]={LastVectors[1][0]-LastVectors[0][0], LastTextures[1][0]-LastTextures[0][0], LastTextures[1][1]-LastTextures[0][1]};
float e1[3]={LastVectors[2][0]-LastVectors[0][0], LastTextures[2][0]-LastTextures[0][0], LastTextures[2][1]-LastTextures[0][1]};

	cp[0]=e0[1]*e1[2]-e0[2]*e1[1];
	cp[1]=e0[2]*e1[0]-e0[0]*e1[2];
	cp[2]=e0[0]*e1[1]-e0[1]*e1[0];

	if(fabs(cp[0])>0.00001f)
	{
		tangent[0]=-cp[1]/cp[0];
		binormal[0]=-cp[2]/cp[0];
	}

	e0[0]=LastVectors[1][1]-LastVectors[0][1];
	e1[0]=LastVectors[2][1]-LastVectors[0][1];

	cp[0]=e0[1]*e1[2]-e0[2]*e1[1];
	cp[1]=e0[2]*e1[0]-e0[0]*e1[2];
	cp[2]=e0[0]*e1[1]-e0[1]*e1[0];

	if(fabs(cp[0])>0.00001f)
	{
		tangent[1]=-cp[1]/cp[0];
		binormal[1]=-cp[2]/cp[0];
	}

	e0[0]=LastVectors[1][2]-LastVectors[0][2];
	e1[0]=LastVectors[2][2]-LastVectors[0][2];

	cp[0]=e0[1]*e1[2]-e0[2]*e1[1];
	cp[1]=e0[2]*e1[0]-e0[0]*e1[2];
	cp[2]=e0[0]*e1[1]-e0[1]*e1[0];

	if(fabs(cp[0])>0.00001f)
	{
		tangent[2]=-cp[1]/cp[0];
		binormal[2]=-cp[2]/cp[0];
	}
}

// -------------------------------------------------------------- //
// gRotateVector: Rotate(don't translate/scale) mat = glGetMatrix style
// -------------------------------------------------------------- //
inline void gRotateVector(const float* mat, float* vec)
{
static float	vec2[3];
	// This is the way MESA worx
	vec2[0]=(vec[0]*mat[0])+(vec[1]*mat[4])+(vec[2]*mat[8]);
	vec2[1]=(vec[0]*mat[1])+(vec[1]*mat[5])+(vec[2]*mat[9]);
	vec2[2]=(vec[0]*mat[2])+(vec[1]*mat[6])+(vec[2]*mat[10]);
	vec[0] = vec2[0];
	vec[1] = vec2[1];
	vec[2] = vec2[2];
}


// -------------------------------------------------------------- //
// gTransformVector: V'=M*V m = glGetMatrix style
// -------------------------------------------------------------- //
inline void gTransformVector(const float* mat, float* vec)
{
static float	vec2[3];
	// This is the way MESA worx
	vec2[0]=(vec[0]*mat[0])+(vec[1]*mat[4])+(vec[2]*mat[8]) +mat[12];
	vec2[1]=(vec[0]*mat[1])+(vec[1]*mat[5])+(vec[2]*mat[9]) +mat[13];
	vec2[2]=(vec[0]*mat[2])+(vec[1]*mat[6])+(vec[2]*mat[10])+mat[14];

	vec[0] = vec2[0];
	vec[1] = vec2[1];
	vec[2] = vec2[2];
}


// --------------------------------------------------------------- //
// base get's shortened, so it points to the perpendicular
// through point pt
// returns whether direction got reverted
// --------------------------------------------------------------- //
inline bool gPerpend(float* base, const float* pt)
{
	// res = uni_base * (base dot pt)
	gUnify(base);
	float dot = gDot(base, pt);
	gMultVertex(base, dot);
	if(dot<0) return false;
return true;
}


// --------------------------------------------------------------- //
// Get distance of 2 vertices
// --------------------------------------------------------------- //
inline float gSQDistOfVertex(const float* v1, const float* v2)
{
float tmp[3];
	tmp[0]=v1[0] - v2[0];
	tmp[1]=v1[1] - v2[1];
	tmp[2]=v1[2] - v2[2];
	return gGetSQLength(tmp);
}

// --------------------------------------------------------------- //
// gSphereInCone - by Gernot Frisch (www.glbasic.com)
// Detects if a sphere (or point: R=0) is inside a cone
// The cone has R=0 in cone_start and cone_angle as an
// opening angle. It's opening towards the vertex: cone_end
// tan_cone_angle: tan(cone_angle)
// center:         The center of the sphere
// R_shpere:       The sphere's radius
// --------------------------------------------------------------- //
inline bool gSphereInCone(const float* cone_start, const float* cone_end, float tan_cone_angle, const float* center, float R_sphere)
{
float bl[3], pt[3], nl[3], R2 = R_sphere*R_sphere;
	// Subtract the cone_start position from all, so the whole system
	// is starting in the origin.
	gDiffVertex(bl, cone_end, cone_start);
	gDiffVertex(pt, center,   cone_start);
	// first get a perpendicular on the cone's center line through the sphere's center 'pt'
	// bl = base line, now shorthened to the perpendicular point
	if(!gPerpend(bl, pt))
	{
		// This thing is somewhere behind the cone
		// see if the radius is greater than the distance
		// of the center to the origin (== start)
		if (R2 >= gGetSQLength(pt))
			return true;
		return false;	
	}

	//                @ center
	//                |
	//                |---
	//              _-^ find this vertex (nl = normal to base line)
	//         __--  |
	// (0,0) @---bl->---@ end
	//         ``--_
	//              `--
	// Get a vertex from that point to the sphere's center
	// but shorten it, so it stops on the surface of the cone
	gDiffVertex(nl, pt, bl);
	gUnify(nl);
	gMultVertex(nl, tan_cone_angle * qsqrt(gGetSQLength(bl))); // 5% inaccuracy -> make sure tan_cone_angle covers this!
	
	// Fine. Now we can test, if the sphere's center is within the cone
	// That is: The distance of 'bl' to 'pt'
	
	// if distance(perpend-point, center) < nl+radius
	if(gSQDistOfVertex(bl, pt) < (gGetSQLength(nl) + R2)) return true;
return false;
}




#endif


