/*
 Copyright (C) 2011 by Stuart Carnie
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 */

#import "iCadeReaderView.h"

@interface iCadeReaderView()

- (void)didEnterBackground;
- (void)didBecomeActive;

@end

@implementation iCadeReaderView

@synthesize iCadeState=_iCadeState;
@synthesize delegate=_delegate;
@synthesize active;

- (id)initWithFrame:(CGRect)frame {
    self = [super initWithFrame:frame];
    inputView = [[UIView alloc] initWithFrame:frame]; /* DaCarSoft: dimensions to same as iCadeView to avoid iOS redimension problem */
    
    if (&UIApplicationDidEnterBackgroundNotification) /* DaCarSoft: iOS 3.0 compatibility */
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(didEnterBackground) name:UIApplicationDidEnterBackgroundNotification object:nil];
    
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(didBecomeActive) name:UIApplicationDidBecomeActiveNotification object:nil];

	_isKbdLayout=0;
    _iCadeAttached = 0;
    return self;
}

- (void)dealloc {
// linker errors with SDK 3.x :(
    if (&UIApplicationDidEnterBackgroundNotification) /* DaCarSoft: iOS 3.0 compatibility */
        [[NSNotificationCenter defaultCenter] removeObserver:self name:UIApplicationDidEnterBackgroundNotification object:nil];
    
    [[NSNotificationCenter defaultCenter] removeObserver:self name:UIApplicationDidBecomeActiveNotification object:nil];
    [super dealloc];
}

- (void)didEnterBackground {
    printf("didEnterBackground\n");
    if (_isActive)
        [self resignFirstResponder];
}

- (void)didBecomeActive {
    printf("didBecomeActive\n");
    if (_isActive)
        [self becomeFirstResponder];
}

- (BOOL)canBecomeFirstResponder { 
    return YES; 
}

- (void)setActive:(BOOL)value {
    printf("setActive\n");
    if (_isActive == value) return;
    _isActive = value;
    if (_isActive) {
        [self becomeFirstResponder];
    } else {
        [self resignFirstResponder];
    }
}

- (UIView*) inputView {
    return inputView;
}

- (void)setDelegate:(id<iCadeEventDelegate>)delegate {
    _delegate = delegate;
    if (!_delegate) return;
    
    _delegateFlags.iCadeStateChanged = [_delegate respondsToSelector:@selector(iCadeStateChanged:)];
    _delegateFlags.iCadeButtonDown = [_delegate respondsToSelector:@selector(iCadeButtonDown:)];
    _delegateFlags.iCadeButtonUp = [_delegate respondsToSelector:@selector(iCadeButtonUp:)];
}

#pragma mark -
#pragma mark UIKeyInput Protocol Methods

- (BOOL)hasText {
    return NO;
}

- (void)insertText:(NSString *)text {
	const char* ON_STATES, *OFF_STATES;

    bool iCadeStateChanged = false;
    char ch = [text characterAtIndex:0];
	/*
     There are three main latin keyboard layouts: QWERTY, QWERTZ and AZERTY  keyboard layouts.
			The QWERTZ layout switches Y - Z
			The AZERTY layout switches A - Q and Z - W


		// iCade events:
		  WE    YT UF IM OG
		AQ  DC
		  XZ    HR JN KP LV

		QUERTZ/AZERTY: W w/o Z (UP,   pressed w/  DOWN, pressed)
		AZERTY:        Q w/o A (LEFT, release w/o LEFT, pressed)
					   Z w/o X (DOWN, release w/o DOWN, pressed)
	*/
	if(_isKbdLayout==0) // still quertz
	{
		if(ch == 'w' && (_iCadeState & iCadeJoystickDown))
		{
			printf("iCade key-layout: quertz\n");
			_iCadeState = iCadeJoystickNone;
			iCadeStateChanged = true;
			_isKbdLayout = 1;
		}
	}
	if(_isKbdLayout<2) // not AZERTY
	{
		if(ch == 'q' && !(_iCadeState & iCadeJoystickLeft))
		{
			printf("iCade key-layout: azerty\n");
			_iCadeState = iCadeJoystickNone;
			iCadeStateChanged = true;
			_isKbdLayout = 2;
		}
	}

	switch(_isKbdLayout) // Space Fractal :: swapped buttons, so 8-bitty and such controllers can been used as default layouyt.
	{
	default:
		ON_STATES  = "wdxahjyuikol";
		OFF_STATES = "eczqrntfmpgv";
		break;
	case 1:
		ON_STATES  = "wdxahjzuikol";
		OFF_STATES = "ecyqrntfmpgv";
		break;
	case 2:
		ON_STATES  = "zdxqhjwuikol";
		OFF_STATES = "ecyarntfmpgv";
		break;
	}

	// GF: store that for inkey$
	_last_character = ch;

	_iCadeAttached = 1; // GF: evidence for a keyboard found

    char *p = strchr(ON_STATES, ch);
    if (p) {
        int index = (int)(p-ON_STATES);
        _iCadeState |= (1 << index);
        iCadeStateChanged = true;
        if (_delegateFlags.iCadeButtonDown) {
            [_delegate iCadeButtonDown:(1 << index)];
        }
    } else {
        p = strchr(OFF_STATES, ch);
        if (p) {
            int index = (int)(p-OFF_STATES);
            _iCadeState &= ~(1 << index);
            iCadeStateChanged = true;
            if (_delegateFlags.iCadeButtonUp) {
                [_delegate iCadeButtonUp:(1 << index)];
            }
        }
    }

    if (iCadeStateChanged && _delegateFlags.iCadeStateChanged) {
        [_delegate iCadeStateChanged:_iCadeState];
    }
    
    static int cycleResponder = 0;
    if (++cycleResponder > 20) {
        // necessary to clear a buffer that accumulates internally
        cycleResponder = 0;
        [self resignFirstResponder];
        [self becomeFirstResponder];
    }
}

- (void)deleteBackward {

	_last_character = '\b'; // just to make that right
}

@end
