/*
Very simple to use network engine that
stores a queue of messages for each client
that can be gained very conveniently.
*/

#include "gffilesystem.h"


#if defined( _DEBUG) // || defined(ANDROID)
	#define d_printf g_debug
#else
	static void d_printf(...) {}
#endif



#define MAX_MESSAGE_STRING	1024
#define DEFAULT_PORT		27910
#define MAX_CLIENTS 128 // how many can connect to the server
#define MAX_PLAYERS 8   // how many players can one program create
#define MAX_RETRY	5*50	// n*50 = n secs for timeout


// msleep ( sleep 1/1000 sec )
#if defined WIN32 || defined WIN64 || defined _WIN32_WCE
	#define WIN32_LEAN_AND_MEAN
	#include <windows.h>
	inline void msleep(int x) {Sleep(x);}
#elif defined(EMSCRIPTEN)

    /* can use usleep if nanosleep is not supported on your platform */
	inline void msleep(int msec){usleep(msec*1000);}

#else
	#include <unistd.h>
	#include <time.h>
	#include <errno.h>
	static void msleep(unsigned int msec)
	{
		struct timespec timeout0;
		struct timespec timeout1;
		struct timespec* tmp;
		struct timespec* t0 = &timeout0;
		struct timespec* t1 = &timeout1;

		t0->tv_sec = msec / 1000;
		t0->tv_nsec = (msec % 1000) * (1000 * 1000);

		while ((nanosleep(t0, t1) == (-1)) && (errno == EINTR))
		{
			tmp = t0;
			t0 = t1;
			t1 = tmp;
		}
	}
	// inline void msleep(int x) {timespec test; test.tv_sec=x/1000; test.tv_nsec = (x%1000)*1000; nanosleep(&test,NULL);}
#endif


#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if defined( _WIN32 ) || defined (_WIN64) || defined( _WIN32_WCE)
	#include <tchar.h>
#endif

#include "gf_sockets.h"


#include "hawkthreads.h"
#pragma comment(lib, "ws2_32.lib")

#define ID_SERVER            0x0000 // The "playerid" of the server
#define ID_ALL_PLAYERS       0xffff // Send to all players
#define ID_NO_GOOD_NAME      0xfffe // a player with this ID will send NLM_WHOAMI to the server and change to ID=ID_GIMME_ID
// #define ID_PLAYER_LOGGED_OFF 0xfffd
#define ID_GIMME_ID          0xfffc // this player sent a NLM_WHOAMI request, and waits for the NLM_YOUARE message
#define ID_INVALID           0xff00 // no valid player ID, if bigger than that
#define VALID_ID(a) ((a)>0 && (a) < ID_INVALID)

#define NLM_WHOAMI	1 // request the playerid that you are (pBuffer = Player structure. address contains port. IP must be filled.) client->server
#define NLM_YOUARE	2 // send the playerid you have server->client. id_recipent is your ID (which is ID_GIMME_ID).
					  // as long as you don't get this, you're not in the server - retry 3x 5secs
#define NLM_SENDMSG	3 // Send a message to id_recipent (ID_ALL_PLAYERS for all)
#define NLM_LOGOFF	4 // This player quits. id_from will be removed

// Queries
#define NLM_QUERY				(100)
#define NLM_QUERY_NUMPLAYERS	(101)	// Get the number of players registered on server
#define NLM_QUERY_PLAYERID		(102)	// Get the playerid of player with index[i] in server list
#define NLM_QUERY_PLAYERNAME	(103)	// Get the name of a player (by id)
#define NLM_QUERY_IS_SERVER_ALIVE (104) // Is the server responding to messages?

#if defined(WIN32) && defined(_DEBUG)
extern "C"
{
int _open_osfhandle(intptr_t, int);
}
#endif


template <class T> class GFDeque
{
public:
	struct item
	{
		item(): pPrev(NULL),pNext(NULL) {}
		item *pPrev, *pNext;
		T data;
	};
	GFDeque(): num_items(0), pFirst(NULL), pLast(NULL) {}
	~GFDeque() {clear();}

	void clear() {while(!empty()) pop_front();}
	void push_back(const T& t)
	{
		item* pT = new item;
		pT->data = t;
		if(pLast)
		{
			pLast->pNext = pT;
			pT->pPrev = pLast;
		}
		pLast = pT;
		if(!pFirst) pFirst=pLast;		
		++num_items;
	}
	void pop_back()
	{
		if(pLast)	{remove(pLast);}
	}
	void push_front(const T& t)
	{
		item* pT = new T;
		pT->data = t;
		if(pFirst)
		{
			pFirst->pPrev = pT;
			pT->pNext = pFirst;
		}
		pFirst = pT;
		if(!pLast) pLast = pFirst;
		++num_items;
	}
	void pop_front()
	{
		if(pFirst)	{remove(pFirst);}
	}
	void remove(item* pWhich) // pointer points to void data afterwards!
	{
		if(!pWhich || !num_items) return;
		item* pBef, *pAfter;
		pBef = pWhich->pPrev;
		pAfter = pWhich->pNext;
		if(pBef) pBef->pNext = pAfter;
		if(pAfter) pAfter->pPrev = pBef;

		if (pWhich == pLast) pLast = pBef;
		if (pWhich == pFirst) pFirst = pAfter;
		delete pWhich;
		--num_items;
	}
	item*	operator[](int i) // access by index [slow]
					{
						if (i<0 || i>=num_items) return NULL;
						item* pI = front();
						for(; i>0; --i)
						{
							pI = pI->pNext;
						}
						return pI;
					}
	item*	front()	{return pFirst;}
	item*	back()	{return pLast;}
	int		size()	{return num_items;}
	bool	empty()	{return pFirst == NULL ? true : false;}
private:
	item	*pFirst, *pLast;
	int num_items;
};

class GFHawkNetwork
{
public:
	// --------------------------------------------------------------- //
	// A message structure
	// --------------------------------------------------------------- //

#pragma pack(push,1) // we need every byte where it is

	class Message // Keep order of member variables (HI-ENDIAN only)
	{
		public:
		unsigned char	whattodo;
		unsigned short	numbytes;
		unsigned short	id_recipent;
		unsigned short	id_from;
		char			pBuffer[MAX_MESSAGE_STRING];

		Message()	{whattodo=0; numbytes=0; id_from=0; id_recipent=0; pBuffer[0]='\0';}
		Message& operator=(const Message& m)
					{memcpy(this, &m, sizeof(Message)); return *this;}
		// flip bytes before sending
		inline void FlipBytes()
		{
			numbytes    = ToLittleEndian(numbytes);
			id_recipent = ToLittleEndian(id_recipent);
			id_from     = ToLittleEndian(id_from);
		}
	};
#pragma pack(pop)

private:
	#define LNM_HEADER_SIZE		(sizeof(Message) - sizeof(char[MAX_MESSAGE_STRING]))
	// --------------------------------------------------------------- //
	// A player structure
	// 144+4(msg) bytes - all 4 byte aligned
	// --------------------------------------------------------------- //
	struct RawPlayer
	{
		NETsocket			sock_pl;	// sock    used to recv from server TCP
		unsigned short		id; // ID_NO_GOOD_NAME if not yet initialized, ID_GIMME_ID sent message to get id, waiting
		char				is_temp_player; // do not list this player in a query to num_players
		char				is_hosted_on_server; // if so, don't send, but put in mailbox
		char				name[128];
		RawPlayer()								{sock_pl=NET_INVALID; name[0]='\0'; name[127]='\0'; id=ID_INVALID; is_temp_player=0; is_hosted_on_server=0;}
		RawPlayer(const RawPlayer& p)			{*this=p;}
		RawPlayer& operator=(const RawPlayer& p){sock_pl = p.sock_pl; id=p.id; strncpy(name, p.name, 127); name[127]='\0'; is_temp_player=p.is_temp_player; return *this;}
	};

	struct Player : public RawPlayer
	{
		GFDeque<Message> messages; // can't copy that to memory buffer on android, thus RawPlayer

		Player()							{}
		Player(const Player& p)				{*this=p;}
		Player& operator=(const Player& p)	{RawPlayer::operator=(p); messages.clear(); return *this;}
		void FlipBytes()
		{
			sock_pl     = (NETsocket)ToLittleEndian((unsigned long)sock_pl);
			id = ToLittleEndian(id);
		}

		int NumBytesAsBuffer() {return 4+2+1+1+128;} // number of bytes if you copy to buffer

		void CopyToBuffer(char* buf)
		{
			this->FlipBytes(); // to NW
			memcpy(buf, &sock_pl, 4);
			// *(int*)buf = sock_pl;
			buf+=4;
			memcpy(buf, &id, 2);
			buf+=2;
			*buf=is_temp_player;
			++buf;
			*buf=is_hosted_on_server;
			++buf;
			strncpy(buf, name, 128);
			buf[127]='\0';
			this->FlipBytes(); // restore
		}
		void CreateFromBuffer(char* buf)
		{
			memcpy(&sock_pl, buf, 4); // direct int copy didn't work on Android (crash)
			buf+=4;
			memcpy(&id, buf, 2);
			buf+=2;
			is_temp_player=*buf;
			++buf;
			is_hosted_on_server=*buf;
			++buf;
			strncpy(name, buf, 128);
			name[127]='\0';
			this->FlipBytes(); // restore
		}
	};

	// --------------------------------------------------------------- //
	// GFNetwork member data
	// --------------------------------------------------------------- //
	unsigned short	newplayernumber; // the server need this information for new players
	unsigned short	queryplayer;     // a temp_player that is used for queries
	NETsocket		sock_sv;		// a socket if you are the server - to talk to clients
	NETsocket		sock_cl;		// a socket for the player-interaction
	GFDeque<Player> client_players, server_players;
	bool		is_server;
	bool		stopped; // want to stop
	bool		running; // thread runs
	bool		allow_joining;
	HTmutex		busymutex; // mutex for member variables that are accessable from outside

	GFDeque<Message>	playermessages;
	GFDeque<Message>	requests;
	GFDeque<unsigned short> removed_players; // a list of players that quit already. Call GetQuitPlayer() to pop from this stack.
	
	#ifdef _DEBUG
	bool		locked;
	#endif
		
public:
	// --------------------------------------------------------------- //
	// C'tor & D'tor
	// --------------------------------------------------------------- //
	GFHawkNetwork()
	{
		stopped=true; running=false; allow_joining=true;
		htMutexInit(&busymutex);
		sock_sv=NET_INVALID; sock_cl=NET_INVALID;
		newplayernumber = ID_SERVER + 101;
		queryplayer=0;
	}
	~GFHawkNetwork()
	{
		for(GFDeque<Player>::item* pIt = client_players.front(); pIt; pIt = pIt->pNext)
		{
			DeletePlayer(pIt->data.id);
		}
		Stop();
		// nlShutdown();
		netShutdown();
		htMutexDestroy(&busymutex);
	}

	// --------------------------------------------------------------- //
	// # # #  COMMON CODE - used by server + client  # # #
	// --------------------------------------------------------------- //
private:


	// Initialize the engine
	bool Init()
				{
					Stop();

					if(sock_sv    !=NET_INVALID) netCloseSocket(sock_sv);
					if(sock_cl    !=NET_INVALID) netCloseSocket(sock_cl);
					client_players.clear();
					server_players.clear();
					removed_players.clear();
					
					sock_sv=NET_INVALID; sock_cl=NET_INVALID;
					newplayernumber = ID_SERVER + 101;
					queryplayer = 0;

					if(!netInit())
					{
						// d_printf("nlInit: %s\n", nlGetErrorStr(nlGetError()) );
						return false;
					}

					//nlHint(NL_REUSE_ADDRESS, NL_TRUE);
					//nlSelectNetwork(NL_IP); // I cannot call this twice - why?
					// d_printf("nlSelectNetwork: %s\n", nlGetErrorStr(nlGetError()) );
					return true;
				}

public:
	// --------------------------------------------------------------- //
	// Set a stop bit to force threads to terminate
	// --------------------------------------------------------------- //
	void	Stop()
				{
					if(Running())
					{
						Lock(this); stopped=true; Unlock(this);
						while(Running())
						{
							msleep(50);
						}
						#ifdef _DEBUG
						d_printf("STOPPED\n");
						#endif
					}
					if(sock_sv!=NET_INVALID)
					{
						netCloseSocket(sock_sv); sock_sv=NET_INVALID;
					}
					if(sock_cl!=NET_INVALID)
					{
						netCloseSocket(sock_cl); sock_cl=NET_INVALID;
					}
					netShutdown();
					
					// reset some of the states
					stopped=true; running=false; allow_joining=true;
					newplayernumber = ID_SERVER + 101;
					queryplayer=0;
				}

	// --------------------------------------------------------------- //
	// Is the thread still running?
	// --------------------------------------------------------------- //
	bool Running()
				{bool rv; Lock(this); rv = running; Unlock(this); return rv;}

	// --------------------------------------------------------------- //
	// Do you allow new players to join?
	// --------------------------------------------------------------- //
	void AllowJoining(bool allow)
				{Lock(this); allow_joining=allow; Unlock(this);}

	// --------------------------------------------------------------- //
	// CreatePlayer - return new player id (0 if failed)
	// if no name is given, it will be a temp_player (not listed)
	// --------------------------------------------------------------- //
	unsigned short CreatePlayer(const char* name=NULL)
				{
					Lock(this);
					Player ply;
					
					if(name)
						strcpy(ply.name, name);
					else
					{
						strcpy(ply.name, "query");
						ply.is_temp_player = 1;
					}

					d_printf("CreatePlayer %s\n", ply.name);


					if(this->is_server)
						ply.is_hosted_on_server=true;
					
					ply.id = ID_NO_GOOD_NAME; // Get an id
					client_players.push_back(ply);
					Unlock(this);

					unsigned short id=ID_INVALID;
					bool do_retry = true;
					for(int retry=0; retry<MAX_RETRY && do_retry; ++retry)
					{
						// Get player id of last added player
						// the main thread will set this if it receives
						// a YOU_ARE message
						Lock(this);
						if (!client_players.empty())
							id = client_players.back()->data.id;
						do_retry = this->running;
						if(!do_retry)
							d_printf("ERROR: not running anymore\n");
						Unlock(this);
						// server gave me an id
						if( VALID_ID(id) )
						{
							#ifdef _DEBUG
							d_printf("client: created player %s with id %d sock %d\n", name, (int)id, (int)sock_cl);
							#endif
							return id; // Yeeha - accepted
						}
						msleep(20);
					}

					d_printf("got no player id back\n");
					
					// OK, we got no valid player ID, so we remove this one silently
					if (!client_players.empty())
					{
						client_players.pop_back();
					}
					return 0; // did not get a nice name from server
				}

	// --------------------------------------------------------------- //
	// DeletePlayer - kill the player at server side
	// --------------------------------------------------------------- //
	void DeletePlayer(unsigned short id)
				{
					if( !VALID_ID(id) ) return;
					Message msg;
					msg.id_from = id;
					msg.id_recipent = ID_SERVER;
					msg.numbytes = 0;
					msg.whattodo = NLM_LOGOFF;
					Send(sock_cl, msg);
				}

	// --------------------------------------------------------------- //
	// Query some information from the server
	// msg.whattodo = iwhat;
	// msg.numbytes=2;
	// *((unsigned short*)msg.pBuffer) = iextra;
	// --------------------------------------------------------------- //
	bool QueryInformation(Message& msg)
				{
					if(queryplayer == 0)
						queryplayer = CreatePlayer();
					if(queryplayer==0)
					{
						#ifdef _DEBUG
						d_printf("Query: [FAIL - no player]\n");
						#endif
						return false;
					}

					msg.id_from = queryplayer;
					msg.id_recipent = ID_SERVER;

					/*
					#ifdef _DEBUG
					#define SW_MSG(a) case a: d_printf("Query: " #a " "); break;
					switch(msg.whattodo)
					{
						SW_MSG(NLM_QUERY_IS_SERVER_ALIVE)
						SW_MSG(NLM_QUERY_NUMPLAYERS)
						SW_MSG(NLM_QUERY_PLAYERID)
						SW_MSG(NLM_QUERY_PLAYERNAME)
					}
					#endif
					*/
					
					requests.clear();
					Send(sock_cl, msg);
					for(int i=0; i<MAX_RETRY; ++i)
					{
						Lock(this);
						if(requests.size())
						{
							msg = requests.front()->data;
							requests.pop_front();
							Unlock(this);

							d_printf("[OK]\n");
							return true;
						}
						Unlock(this);
						msleep(20);
					}

					d_printf("[FAIL]\n");
					return false;
				}
	
	// --------------------------------------------------------------- //
	// Receive a message structure for the socket
	// >0: OK
	//  0: disconnected
	// -1: error while reading
	// -2: error: stream busy
	// --------------------------------------------------------------- //
	static int Recv(NETsocket sock, Message& msg)
				{
					msg.numbytes=0;
					int nread;
					// will always read complete package at once!!
					// nread = nlRead(sock, &msg, sizeof(msg));
					nread = netRecv(sock, (char*)&msg, LNM_HEADER_SIZE);
					msg.FlipBytes();
					if(nread <= 0) return nread; // just nothing there, yet. Or pending
					
					// we've got the header, we need the rest, too
					if(msg.numbytes)
					{
						nread = netRecv(sock, msg.pBuffer, msg.numbytes, true); // force to read it all
					}
					
					if(nread<=0) return nread;
					// d_printf("Recv from %d to %d\n", msg.id_from, msg.id_recipent);

					// add a terminating null character
					// does not matter for data other than strings
					msg.pBuffer[msg.numbytes] = '\0';
					return true;
				}
	
	// --------------------------------------------------------------- //
	// Send a message structure with the socket
	// --------------------------------------------------------------- //
	static bool Send(NETsocket sock, Message& msg)
				{
					char* pMsg = (char*) &msg;
					int nwrite, ntotal = msg.numbytes + LNM_HEADER_SIZE;

					msg.FlipBytes();
					nwrite = netSend(sock, (const char*)pMsg, ntotal);
					if(nwrite <=0) return false;

					msg.FlipBytes();
					return true;
				}

	// --------------------------------------------------------------- //
	// # # #  SERVER CODE  # # #
	// --------------------------------------------------------------- //
public:
				int GetLocalIPs(char* pNames, int max_len)
				{
					netInit();
					int count=0;
					const char* pAddress = netGetLocalIPs(&count, "|");
					if(!pAddress) return 0;

					strncpy(pNames, pAddress, max_len);
					pNames[max_len] = '\0';
					char* last = strrchr(pNames, '|');
					if(last) *last='\0';
					netShutdown();
					return count;
				}

	// --------------------------------------------------------------- //
	// Start up the server - *this is the server
	// --------------------------------------------------------------- //


				bool StartServer(unsigned short in_port_tcp=DEFAULT_PORT)
				{
					if(!Init()) return false;

					// create the server socket
					sock_sv = netListen(in_port_tcp);
					if(sock_sv == NET_INVALID)
					{
						return false;
					}

					// Now make a client for the server as well - the server can create players!
					sock_cl = netConnect( netAddress(NULL, in_port_tcp) ); // #NEED was 127.0.0.1 before
					if(sock_cl == NET_INVALID) return false;

					// the server is UP now.
					// start the server thread
					is_server=true;
					running=true;
					stopped=false;
					htThreadCreate(mainLoop, (void *)this, HT_FALSE);
					return true;
				}

	// --------------------------------------------------------------- //
	// Get the socket used to send from server to this player
	// --------------------------------------------------------------- //
	static NETsocket GetSockFromServerPlayer(GFHawkNetwork* pEngine, unsigned short id)
				{
					for(GFDeque<Player>::item* pPly = pEngine->server_players.front(); pPly; pPly = pPly->pNext)
					{
						if (pPly->data.id == id)
						{
							return pPly->data.sock_pl;
						}
					}
					return NET_INVALID;
				}

	// --------------------------------------------------------------- //
	// The Main Server Thread
	// --------------------------------------------------------------- //
	static void *mainLoop(void *pArg)
	{
		GFHawkNetwork* pEngine = (GFHawkNetwork*) pArg;
		NETsocket	clientsocks		[MAX_CLIENTS];
		int			clientnum = 0;
		Message		msg;

		for(int i=0; i<MAX_CLIENTS; ++i)
		{
			clientsocks[i]     = NET_INVALID;
		}

		// group = nlGroupCreate();
		Lock(pEngine);
			pEngine->stopped=false; pEngine->running=true;
		Unlock(pEngine);
		bool bThereWasSomeAction;
		for(;;)
		{
			bThereWasSomeAction = false; // decide to go in msleep mode

			// Want to stop it?
			Lock(pEngine);
			if(pEngine->stopped)
			{
				Unlock(pEngine);
				break;
			}
			
			// 1. look, if you requested to create/delete players
			//    and send these special messages to the server
			for(GFDeque<Player>::item* pPly = pEngine->client_players.front(); pPly; pPly = pPly->pNext)
			{
				switch(pPly->data.id)
				{
				case ID_NO_GOOD_NAME: // No good name, yet - send server a message to get one
					bThereWasSomeAction=true;
					#ifdef _DEBUG
					d_printf("WHOAMI: player '%s' needs index and gets: %d (ID_GIMME_ID)\n", pPly->data.name, ID_GIMME_ID);
					#endif
					msg.numbytes = pPly->data.NumBytesAsBuffer();

					pPly->data.CopyToBuffer(msg.pBuffer);
					// memcpy(msg.pBuffer, &pPly->data, sizeof(Player));

					msg.whattodo = NLM_WHOAMI;
					msg.id_recipent = ID_SERVER;
					{
						Unlock(pEngine);
						Send(pEngine->sock_cl, msg);
						Lock(pEngine);
					}
					pPly->data.id = ID_GIMME_ID; // waiting for an id
					break;
				default: // that's a good player - don't do anything
					break;
				}
			}
			Unlock(pEngine);
			
			// 2. Get the messages sent to the client socket (TCP)
			int imsg=0, irecv;
			for(;;)
			{
				irecv = pEngine->Recv(pEngine->sock_cl, msg);
				if(irecv<=0) break;
				++imsg;
				Lock(pEngine);
				pEngine->playermessages.push_back(msg);
				d_printf("recv: %s\n", msg.pBuffer);
				Unlock(pEngine);
				bThereWasSomeAction=true;
			}
			// check if connection died. Then close thread
			if(irecv == 0)
			{
				break;
			}
			
			#ifdef _DEBUG
			if(imsg)
				d_printf("%d client msgs received\n", imsg);
			#endif


			// 3. Get messages for the server (if we are a server)
			if (pEngine->sock_sv != NET_INVALID)
			{
				int count;
				NETsocket	newsock;
				// check for a new client
				bool bjoin;
				Lock(pEngine);
				bjoin = pEngine->allow_joining;
				Unlock(pEngine);

				if(bjoin)
				{
					NETaddress newaddress;
					// accept a connection for a TCP server socket
					newsock = netAccept(pEngine->sock_sv, &newaddress);

					if(newsock != NET_INVALID)
					{
						// find a free socket
						clientsocks    [clientnum] = newsock;
						
						clientnum++;
						d_printf("a new client connected: Sock=%d, clientnum=%d\n", newsock, clientnum);
						bThereWasSomeAction=true;
					}
				}
				
				count = clientnum;
				// loop through the clients and read the packets
				for(int i=0; i<count; i++)
				{
					// Message msg;
					int	good;
					// Receive a package - TCP
					// the udp packages per player are sent to the sock_cl or sock_sv
					good = pEngine->Recv(clientsocks[i], msg);
					if(good>0) // good read
					{
						Lock(pEngine);
						// SET_FROM_ID <-- find this
						// Very special case - we don't know the id/port/socket for this player, yet
						if (msg.whattodo==NLM_WHOAMI)
						{
							#ifdef _DEBUG
							d_printf("recv WHOAMI for index %d, id=%d\n", i, (int)msg.id_from);
							#endif
							msg.id_from = (unsigned short)i;
						}
						// put message to post box
						pEngine->playermessages.push_back(msg);
						Unlock(pEngine);
						bThereWasSomeAction=true;
					}
					else
					{
						// might be a pending, non-blocking UDP connection. No harm done.
						
						// uh-oh. Disconnect information
						if(good == 0)
						{
							d_printf("server: client disconnected %d\n", i);

							for(GFDeque<Player>::item* pPly = pEngine->server_players.front(); pPly; pPly = pPly->pNext)
							{
								if(pPly->data.sock_pl == clientsocks[i])
								{
									d_printf("Removed player: %d, %s\n", (int)pPly->data.id, pPly->data.name);

									if(pEngine->is_server) pEngine->removed_players.push_back(pPly->data.id);

									pEngine->server_players.remove(pPly);
									break;
								}
							}
							--clientnum;
							if(clientnum < 0) clientnum = 0;

							netCloseSocket(clientsocks[i]);
							clientsocks[i]     = clientsocks[clientnum-1]; // remove last to this place
							clientsocks[clientnum] = NET_INVALID;
							bThereWasSomeAction=true;
						}
					}
				}
			} // endif

			// 4. Now put NLM_SENDMSG messages for the players in their post-box
			// and remove messages that could be delivered perfectly
			GFDeque< GFDeque<Message>::item* > pDel; pDel.clear();
			Lock(pEngine);
			for(GFDeque<Message>::item* pItem = pEngine->playermessages.front(); pItem; pItem=pItem->pNext)
			{
				d_printf("in %s\n", pItem->data.pBuffer);
				if(!pItem) break;
				if(pItem->data.whattodo != NLM_SENDMSG) continue;
				
				// check my local clients and put their msgs in mailbox
				for(GFDeque<Player>::item* pPly = pEngine->client_players.front(); pPly; pPly = pPly->pNext)
				{
					if (	pItem->data.id_recipent == pPly->data.id)
					{
						d_printf("client: internally delivered message '%s' for %s from %d\n", pItem->data.pBuffer, pPly->data.name, (int)pItem->data.id_from);

						pPly->data.messages.push_back(pItem->data);
						// Delete if only for this player
						if(pItem->data.id_recipent!=ID_ALL_PLAYERS)
						{
							pDel.push_back(pItem);
							break;
						}
					}
				}
			}
			for(int i=0; i<pDel.size(); ++i)
			{
				GFDeque<Message>::item* pItem = pDel[i]->data;
				pEngine->playermessages.remove(pItem);
				d_printf("Removing internal msg\n");
			}
			
			
			// 5. process some of the messies
			//    for player login/logout
			// GFDeque<Message> internal_postbox; internal_postbox.clear();

			// This is a server only message
			if(pEngine->sock_sv!=NET_INVALID)
			{
				if(pEngine->playermessages.size())
					bThereWasSomeAction=true;

				// a post box for messages that's sent to a server
				// player - no physical send, then
				for(GFDeque<Message>::item* pItem = pEngine->playermessages.front(); pItem; pItem=pItem->pNext)
				{
				cont_message1:
					msg.id_from=ID_ALL_PLAYERS; msg.id_recipent = ID_ALL_PLAYERS; msg.numbytes=0; msg.whattodo = 0xff;

					// Forward a message to a/all clients
					if(pItem->data.whattodo == NLM_SENDMSG)
					{
						msg = pItem->data;
						// send to all players
						if(msg.id_recipent == ID_ALL_PLAYERS)
						{
							for(GFDeque<Player>::item* pPly = pEngine->server_players.front(); pPly; pPly = pPly->pNext)
							{
								// Perform a physical send
								Unlock(pEngine);
								if (!pPly->data.is_temp_player && !pPly->data.is_hosted_on_server)
								{
									msg.id_recipent = pPly->data.id; // this time give a unique id
									if(msg.id_recipent!=msg.id_from)
									{
										Send(pPly->data.sock_pl, msg);
									}
								}
								Lock(pEngine);
								// #NEED - if failed, kill player?
							}
						}
						else // or individual player
						{
							NETsocket sock = GetSockFromServerPlayer(pEngine, pItem->data.id_recipent);
							if(sock!=NET_INVALID)
							{
								Unlock(pEngine);
								Send(sock, msg);
								Lock(pEngine);
							}
						}
						// remove this one
						goto remove_message1;
					}


					// Next messages are requests for the SERVER ONLY
					if(pItem->data.id_recipent != ID_SERVER) continue;

					switch(pItem->data.whattodo)
					{
						case NLM_WHOAMI:
						{
							if(pEngine->allow_joining)
							{
								Player Pl;
								Pl.CreateFromBuffer(pItem->data.pBuffer);

								d_printf("copy done\n");
								d_printf("id %d\n", (int)Pl.id);
								d_printf("sock %d\n", (int)Pl.sock_pl);
								d_printf("name: \"%s\"\n", Pl.name);
								
								// SET_FROM_ID <-- find this
								// Very special case - has been set before explicitly
								
								// the id_from is the index in the clientsocks[] array.
								// with this we can assign the TCP socket.
								Pl.sock_pl = clientsocks[pItem->data.id_from];

								pEngine->server_players.push_back(Pl);

								// pointer to player on server is player socket on server
								pEngine->server_players.back()->data.id = pEngine->newplayernumber++;
								#ifdef _DEBUG
								d_printf("server: created player %d (%s)\n", (int)pEngine->server_players.back()->data.id, pEngine->server_players.back()->data.name);
								#endif
								// Now send a message to inform the player about it's id
								msg.numbytes = 0;
								msg.whattodo = NLM_YOUARE;
								msg.id_from = ID_SERVER;
								msg.id_recipent = pEngine->server_players.back()->data.id;
								Unlock(pEngine);
								Send(Pl.sock_pl, msg);
								Lock(pEngine);
							}
							// remove this one
							goto remove_message1;
							break;
						}
						case NLM_LOGOFF:
						{
							#ifdef _DEBUG
							d_printf("Server got a logoff message from %d\n", (int)pItem->data.id_from);
							#endif
							redo_logoff:
							for(GFDeque<Player>::item* pPly = pEngine->server_players.front(); pPly; pPly = pPly->pNext)
							{
								if (pPly->data.id == pItem->data.id_from) // that's the one to remove
								{
									#ifdef _DEBUG
									d_printf("server: logoff player %d %s\n", (int)pPly->data.id, pPly->data.name);
									#endif

									if(pEngine->is_server) pEngine->removed_players.push_back(pPly->data.id);

									// remove the player
									// GFDeque<Player>::item* pPrev = pPly->pPrev;
									pEngine->server_players.remove(pPly);
									// if(!pPrev) pPrev = pEngine->server_players.front();
									// pPly = pPrev;
									goto redo_logoff;
								}
							}
							// remove this one
							goto remove_message1;
							break;
						}
						case NLM_QUERY_IS_SERVER_ALIVE:
						{
							#ifdef _DEBUG
							d_printf("query IS_ALIVE id: %d\n", (int)pItem->data.id_from);
							#endif
							msg.numbytes = 2;
							*((unsigned short*)msg.pBuffer) = ToLittleEndian((unsigned short)1);
							msg.id_recipent = pItem->data.id_from;
							msg.id_from=ID_SERVER;
							msg.whattodo=NLM_QUERY_IS_SERVER_ALIVE;
							NETsocket sock = GetSockFromServerPlayer(pEngine, msg.id_recipent);

							if(sock!=NET_INVALID)
							{
								Unlock(pEngine);
								Send(sock, msg);
								Lock(pEngine);
							}
							// remove this one
							goto remove_message1;
							break;
						}
						case NLM_QUERY_NUMPLAYERS:
						{
							NETsocket sock = GetSockFromServerPlayer(pEngine, pItem->data.id_from);
							#ifdef _DEBUG
							d_printf("query NUM_PLAYERS id: %d, sock: %d\n", (int)pItem->data.id_from, sock);
							#endif
							msg.numbytes = 2;
							unsigned short num_pls = 0;
							for(int i=0; i<pEngine->server_players.size(); ++i)
							{
								if(!pEngine->server_players[i]->data.is_temp_player && VALID_ID(pEngine->server_players[i]->data.id) )
								{
									++num_pls;
								}
							}
							#ifdef _DEBUG
							d_printf("    will send (%d)\n", (int)num_pls);
							#endif
							
							*((unsigned short*)msg.pBuffer) = ToLittleEndian(num_pls);
							
							msg.id_recipent = pItem->data.id_from;
							msg.id_from=ID_SERVER;
							msg.whattodo=NLM_QUERY_NUMPLAYERS;
							Unlock(pEngine);
							if(sock!=NET_INVALID)
								Send(sock, msg);
							Lock(pEngine);
							// remove this one
							goto remove_message1;
							break;
						}
						case NLM_QUERY_PLAYERID:
						{
							int seekfor = FromLittleEndian(*((unsigned short*)pItem->data.pBuffer));
							#ifdef _DEBUG
							d_printf("query PLAYER_ID [%d]\n", seekfor);
							#endif
							// skip temp players
							for(int i=0; i<pEngine->server_players.size(); ++i)
							{
								if(pEngine->server_players[i]->data.is_temp_player)
									continue;
								if(--seekfor < 0)
								{
									seekfor = i;
									break;
								}
							}

							msg.numbytes = 2;
							*((unsigned short*)msg.pBuffer) = 0;
							if(seekfor>=0 && seekfor<pEngine->server_players.size())
							{
								GFDeque<Player>::item* pPly = pEngine->server_players[seekfor];
								if(pPly)
								{
									*((unsigned short*)msg.pBuffer) = ToLittleEndian(pPly->data.id);
									#ifdef _DEBUG
									d_printf("server: found player[%d].id=%d (%s)\n", seekfor, (int)pPly->data.id, pPly->data.name);
									#endif
								}
							}
							msg.id_recipent = pItem->data.id_from;
							msg.id_from=ID_SERVER;
							msg.whattodo=NLM_QUERY_PLAYERID;
							NETsocket sock = GetSockFromServerPlayer(pEngine, pItem->data.id_from);
							if(sock!=NET_INVALID)
							{
								Unlock(pEngine);
								Send(sock, msg);
								Lock(pEngine);
							}
							// remove this one
							goto remove_message1;
							break;
						}
						case NLM_QUERY_PLAYERNAME:
						{
							unsigned short seekfor = FromLittleEndian(*((unsigned short*)pItem->data.pBuffer));
							msg.numbytes = 0;
							msg.pBuffer[0] = '\0';
							for(GFDeque<Player>::item* pPly = pEngine->server_players.front(); pPly; pPly = pPly->pNext)
							{
								if (pPly->data.id == seekfor)
								{
									d_printf("serverplayer's name: %d = %s\n", (int)seekfor, pPly->data.name);
									strcpy(msg.pBuffer, pPly->data.name);
									break;
								}
							}
							
							// player id not found
							if(msg.pBuffer[0] == '\0')
							{
								strcpy(msg.pBuffer, "NO_DATA");
								d_printf("### ERROR: Can't find playername for %d\nListing...", (int)seekfor);
								for(GFDeque<Player>::item* pPly = pEngine->server_players.front(); pPly; pPly = pPly->pNext)
								{
									d_printf("%4d %s (temp=%d)\n", (int)pPly->data.id, pPly->data.name, (int)pPly->data.is_temp_player);
								}
							}

							msg.numbytes = (unsigned short)strlen(msg.pBuffer);
							msg.id_recipent = pItem->data.id_from;
							msg.id_from=ID_SERVER;
							msg.whattodo=NLM_QUERY_PLAYERNAME;
							NETsocket sock = GetSockFromServerPlayer(pEngine, pItem->data.id_from);
							if(sock!=NET_INVALID)
							{
								Unlock(pEngine);
								Send(sock, msg);
								Lock(pEngine);
							}
							// remove this one
							goto remove_message1;
							break;
						}
					}
					continue;
				remove_message1:
						Lock(pEngine);
						GFDeque<Message>::item* pDelItem = pItem;
						pItem = pItem->pNext;
						pEngine->playermessages.remove(pDelItem); pDelItem=NULL;
						Unlock(pEngine);
						if(pItem) goto cont_message1;
						else break;
				}
			}
			
			// Special messages for the client, from the server
			for(GFDeque<Message>::item* pItem = pEngine->playermessages.front(); pItem; pItem=pItem->pNext)
			{
			cont_message2:
				if(pItem->data.id_from != ID_SERVER) continue;
				// is it a query? store it
				if(pItem->data.whattodo>=NLM_QUERY)
				{
						pEngine->requests.push_back(msg);
						goto remove_message2;
				}
				switch(pItem->data.whattodo)
				{
					case NLM_YOUARE:
					{
						for(GFDeque<Player>::item* pPly = pEngine->client_players.front(); pPly; pPly = pPly->pNext)
						{
							if (pPly->data.id == ID_GIMME_ID) // This one requested an id and is waiting for it now
							{
								pPly->data.id = pItem->data.id_recipent;
								#ifdef _DEBUG
								d_printf("client: got player id %d for %s\n", (int)pPly->data.id, pPly->data.name);
								#endif
								break;
							}
						}
						// remove this one
						goto remove_message2;
					}
				}
				continue;
			remove_message2:
					GFDeque<Message>::item* pDelItem = pItem;
					pItem = pItem->pNext;
					pEngine->playermessages.remove(pDelItem);
					if(pItem) goto cont_message2;
					else break;
			}
			
			// 6. Drop all messages for players that cannot be delivered
			#ifdef _DEBUG
			if(pEngine->playermessages.size())
			{
				d_printf("dropped %d messages\n", pEngine->playermessages.size());
				for(GFDeque<Message>::item* pItem = pEngine->playermessages.front(); pItem; pItem=pItem->pNext)
				{
					d_printf ("Act: %d, recip: %d\n", (int)pItem->data.whattodo, (int)pItem->data.id_recipent);
				}
			}
			#endif
			pEngine->playermessages.clear();
			
			// 7. finally apply internal messages for next time
			/*
			for(int i=0; i<internal_postbox.size(); ++i)
			{
				pEngine->playermessages.push_back(internal_postbox.back()->data);
				internal_postbox.pop_back();
			}
			*/
			
			Unlock(pEngine);
			
			
			// This gives the other processes more time, right?
			// Should I set this higher, or will I loose packages (with UDP)
			if(!bThereWasSomeAction)
				msleep(10);
		}
		Lock(pEngine);
		
		for(int i=0; i<MAX_CLIENTS; ++i)
		{
			if(clientsocks[i] != NET_INVALID)
				netCloseSocket(clientsocks[i]);
			clientsocks[i] = NET_INVALID;
		}

		pEngine->running=false;
		pEngine->stopped=true;
		Unlock(pEngine);
		
		pEngine->Stop(); // close sockets

		#ifdef _DEBUG
		d_printf("} thread - shut DOWN\n");
		#endif
		return NULL;
	}

	// --------------------------------------------------------------- //
	// # # #  CLIENT CODE  # # #
	// --------------------------------------------------------------- //
public:
	// --------------------------------------------------------------- //
	// Connect the Client to the Server - *this is the client
	// --------------------------------------------------------------- //
	bool ConnectClient(const char* server_url, unsigned short in_port_tcp=DEFAULT_PORT)
	{
		if (!Init()) return false;

		NETaddress server = netAddress(server_url, in_port_tcp);
		if(server == NET_ADDR_INVALID) return false;

		sock_cl = netConnect(server);
		if(sock_cl == NET_INVALID)
			return false;


		// Nice, the client could connect. Startup the message thread
		is_server=false;
		running=true;
		stopped=false;
		htThreadCreate(mainLoop, (void *)this, HT_FALSE);
		
		Message msg;
			msg.numbytes=0;
			msg.whattodo = NLM_QUERY_IS_SERVER_ALIVE;
		bool good = this->QueryInformation(msg);
		if(!good)
		{
			d_printf("Server is dead, shutting down\n");
			Stop();
			return false;
		}
		return true;
	}

	// --------------------------------------------------------------- //
	// Get the number of messages received in the backgorund
	// --------------------------------------------------------------- //
	int  NumWaitingMessages(unsigned short id_player)
	{
		int num = 0;
		Lock(this);
		for(GFDeque<Player>::item* pIt = client_players.front(); pIt; pIt = pIt->pNext)
		{
			if (id_player == pIt->data.id)
			{
				num = pIt->data.messages.size();
				break;
			}
		}
		Unlock(this);
		return num;
	}


	// --------------------------------------------------------------- //
	// return player id that just quit, or 0 if none quit, yet
	// --------------------------------------------------------------- //
	unsigned short GetQuitPlayer()
	{
		unsigned short rv=0;
		Lock(this);
		if(this->is_server && removed_players.size())
		{
			rv = removed_players.back()->data;
			removed_players.pop_back();
		}
		Unlock(this);
		return rv;
	}
	

	// --------------------------------------------------------------- //
	// Get the number of messages received in the backgorund
	// --------------------------------------------------------------- //
	unsigned short  NumPlayers()
	{
		if(!Running()) return 0;

		Message msg;
		msg.numbytes=0;
		msg.whattodo = NLM_QUERY_NUMPLAYERS;
		if (QueryInformation(msg))
		{
			unsigned short np = FromLittleEndian(*((unsigned short*)msg.pBuffer));
			d_printf("num players: %s = %d\n", msg.pBuffer, (int)np);
			return np;
		}
		return 0;
	}
	
	// --------------------------------------------------------------- //
	// Get the name of player by index in server list (numplayers)
	// --------------------------------------------------------------- //
	unsigned short  GetPlayerIdByIndex(unsigned short nplayer)
	{
		Message msg;
		msg.numbytes=2;
		*((unsigned short*)msg.pBuffer) = ToLittleEndian(nplayer);
		#ifdef _DEBUG
		d_printf("client: request id for %d\n", (int)nplayer);
		#endif
		msg.whattodo = NLM_QUERY_PLAYERID;
		if (QueryInformation(msg))
		{
			return FromLittleEndian(*((unsigned short*)msg.pBuffer));
		}
		return 0;
	}
	
	// --------------------------------------------------------------- //
	// Get the name of player by id
	// --------------------------------------------------------------- //
	bool  GetPlayerName(unsigned short id_player, char* pName)
	{
		Message msg;
		msg.numbytes=2;
		d_printf("Get Player Name for %d\n", (int)id_player);
		*((unsigned short*)msg.pBuffer) = ToLittleEndian(id_player);
		msg.whattodo = NLM_QUERY_PLAYERNAME;
		if (QueryInformation(msg))
		{
			strcpy(pName, msg.pBuffer);
			return true;
		}
		return false;
	}

	// --------------------------------------------------------------- //
	// Get the oldest message on the queue
	// --------------------------------------------------------------- //
	bool PopMessage(unsigned short id_player, Message& Msg)
	{
		bool rv=false;
		Lock(this);
		for(GFDeque<Player>::item* pIt = client_players.front(); pIt; pIt = pIt->pNext)
		{
			if (id_player == pIt->data.id)
			{
				if(pIt->data.messages.size())
				{
					Msg = pIt->data.messages.front()->data;
					pIt->data.messages.pop_front();
					rv=true;
				}
				break;
			}
		}
		Unlock(this);
		return rv; // #need: stack
	}

	// --------------------------------------------------------------- //
	// Make a Message structure from a string and send it to the server
	// --------------------------------------------------------------- //
	bool SendASCIIMessage(unsigned short player_from, unsigned short recipent, const char* message)
	{
		Message msg;
		msg.whattodo = NLM_SENDMSG;
		msg.numbytes = (unsigned short)strlen(message);
		strcpy(msg.pBuffer, message);
		msg.id_from = player_from;
		msg.id_recipent = recipent;
		return Send(sock_cl, msg);
	}


	// --------------------------------------------------------------- //
	// WGet - download a file from internet
	// localfile - file to write, NULL if no writing
	// contents  - contents buffer, NULL if only to file
	// --------------------------------------------------------------- //
	bool WGet(const char* cserver, const char* curl, unsigned short port, const char* localfile, char* contents=NULL, long max_contents_length=0, int timeout_ms=5000)
	{
		int		count, total = 0, buflen = 0;
		int		crfound = 0;
		int		lffound = 0;
		char*	pAppend=contents;
		char*   buffer = NULL;
		
		// default http port
		if (port == 0) port = 80;
		
		GFFileSystem file;
		if(localfile)
		{
			if(!file.OpenOut(localfile))
			{
				return false;
			}
		}
		if(!netInit())
		{
			return false;
		}
		NETaddress addr = netAddress(cserver, port);
		if(addr == NET_ADDR_INVALID)
		{
			return false;
		}
		
		NETsocket sock = netConnect(addr, timeout_ms);
		if(sock == NET_INVALID)
		{
			goto failed;
		}
		
		// blocking mode
		netSetBlocking(sock, false);
		
		// i/o buffer
		buflen = max(1024*16, 100 + strlen(cserver) + strlen(curl));
		buffer = new char [ buflen ];
		sprintf(buffer, "GET %s HTTP/1.1\r\nHost: %s\r\nAccept: */*\r\n"
						// "Session-Id: SID:ANON:w3.org:j6oAgxCWZh/CD723LGeXlf-01:034\r\n"
						"Connection: Close\r\n"
						"User-Agent: GLBasic\r\n\r\n",curl ,cserver);
		
		if(netSend(sock, buffer, (int)strlen(buffer), NULL) <= 0)
		{
			goto failed;
		}

		// receive the file and write it locally
		for(;;)
		{
			count = netRecv(sock, buffer, buflen-1);
			if (count == -2)
			{
				msleep(5);
				continue;
			}
			if(count == 0) break; // end of transmission
			if(count < 0)
				goto failed;

			total += count;
			// parse out the HTTP header
			if(lffound < 2)
			{
				int i;
				for(i=0;i<count;i++)
				{
					if(buffer[i] == '\r')
					{
						crfound++;
					}
					else
					{
						if(buffer[i] == '\n')
						{
							lffound++;
						}
						else
						{
							// reset the CR and LF counters back to 0
							crfound = lffound = 0;
						}
					}
					if(lffound == 2)
					{
						// i points to the second LF
						// NUL terminate the header string and print it out
						buffer[i] = buffer[i-1] = 0x0;

						// write out the rest to the file
						count = count-i-1;

						if(localfile)
							file.Write(&buffer[i+1], count);

						if(pAppend && count>0 && max_contents_length>1)
						{
							int ncopy = min(count, (int)max_contents_length-1);
							strncpy(pAppend, buffer+i+1, ncopy );
							pAppend += ncopy;
							*pAppend = '\0';
							max_contents_length-=ncopy;
						}
						break;
					}
				}
			}
			else
			{
				if(localfile)
					file.Write(buffer, count);

				if(pAppend && count>0 && max_contents_length>1)
				{
					int ncopy = min(count, (int)max_contents_length-1);
					strncpy(pAppend, buffer, ncopy );
					pAppend += ncopy;
					*pAppend = '\0';
					max_contents_length-=ncopy;
					if(max_contents_length<=1) break; // no more space to append
				}
			}
		}
		if(localfile)
			file.Close();
		netCloseSocket(sock);

		if(buffer) delete[] buffer;
		return true;

	failed:
		if(sock != NET_INVALID) netCloseSocket(sock);
		
		int err = netLastErrNum();
		netShutdown();
		netStoErr(err);

		if(buffer) delete[] buffer;
		return false;
	}


private:
	static void Lock(GFHawkNetwork* pEngine)
	{
		htMutexLock(&pEngine->busymutex);
		#ifdef _DEBUG
		pEngine->locked=true;
		#endif
	}
	static void Unlock(GFHawkNetwork* pEngine)
	{
		#ifdef _DEBUG
		pEngine->locked=false;
		#endif
		htMutexUnlock(&pEngine->busymutex);
	}
};


#ifdef WANT_HAWK_MAIN
// --------------------------------------------------------------- //
// MAIN
// --------------------------------------------------------------- //
int main(int argc, char **argv)
{
GFHawkNetwork serv; //, client;
	// Init both: sever + client
//	printf("Init: %d\n", serv.Init());
//	printf("Init: %d\n", client.Init());
int IS_SERVER=0;
unsigned short myplayer;
	if(serv.ConnectClient(DEFAULT_PORT, "127.0.0.1"))
	{
		printf("**************\nClient: 1\n");
		myplayer=serv.CreatePlayer("Clark");
		printf("Clark: %d\n", myplayer);
	}
	else
	{
		IS_SERVER=1;
		// Start the server
		printf("**************\nServer: %d\n", serv.StartServer());
		myplayer=serv.CreatePlayer("Sam");
		printf("Sam: %d\n", myplayer);
	}

unsigned short theother=0;

	int msg=0;
	while(true)
	{
		msg++;
		char ms[1024];
		if(theother)
		{
			if( IS_SERVER)
			{
				printf("Sending\n");
				sprintf(ms, "Sam 2 Clark %d.", msg);
				serv.SendASCIIMessage(myplayer, theother, ms);
			}
			else
			{
				int num = serv.NumPlayers();
				for (int n=0; n<num; ++n)
				{
					unsigned short id = serv.GetPlayerIdByIndex(n);
					if (id!=myplayer) theother=id;
					serv.GetPlayerName(id, ms);
					printf ("    Player: %d: id=%d, %s\n", n, id, ms);
				}
				printf("Sending\n");
				sprintf(ms, "Clark 2 Sam %d.", msg);
				serv.SendASCIIMessage(myplayer, theother, ms);
			}
			while(serv.NumWaitingMessages(myplayer))
			{
				GFHawkNetwork::Message messy;
				serv.PopMessage(myplayer, messy);
				printf("Got Mail: %s\n", messy.pBuffer);
			}
		}
		else // try finding out who the other is...
		{
			int num = serv.NumPlayers();
			printf("List: %d players in game\n", num);
			for(int i=0; i<num; ++i)
			{
				unsigned short id = serv.GetPlayerIdByIndex(i);
				char name[1024]="\0";
				serv.GetPlayerName(id, name);
				printf ("index %4d has ID %5d (%s)\n", i, id, name);
				if(id!=myplayer) {theother=id; break;}
			}

			printf("The other player = %d\n", theother);
		}
		printf("Press enter\n");
		getchar();
	}
	
	serv.Stop();	
	return 0;
}
#endif


