#ifndef PLANNER_H
#define PLANNER_H
/**
  @class Planner
  @brief Interface for path-planning algorithms.

  This provides a general interface for use with various path-finding
  algorithms.  Common things like initializing the map are handled here,
  while the actual path-finding algorithms are implemented in classes derived
  from this.

  Copyright (C) 2003 Ted Cipicchio <ted@thereisnospork.com>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to
  deal in the Software without restriction, including without limitation the
  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  IN THE SOFTWARE.

  @date  Created: Thu Mar 20 2003
*/

#include "node.h"
#include "fixeddeque.h"
#include "nodehashtable.h"

/// Wall identifier.
#define PLANNER_WALL -1.0f

//----------------------------------------------------------------------------
class Planner
{
public:
	/// Planner run description.
	/**
	  This is used to describe how the planner should function when finding a
	  path.

	  @note  The @c width and @c height properties are ignored by createMap()
	         when creating the map to plan on.  createMap() uses the
	         dimensions from the image passed itself.  @c width and @c height
	         are provided for getting the dimensions back from the planner at
	         a later time.
	*/
	struct Description
	{
		/// Start x-coordinate.
		int startX;
		/// Start y-coordinate.
		int startY;

		/// End x-coordinate.
		int endX;
		/// End y-coordinate.
		int endY;

		/// Use increased cost for diagonal movement (~1.4142x).
		bool bDiagonalCost;
		bool can_walk_diagonally;

		/// Threshold value for getting walls from image (sum of RGB values).
		// int wallThreshold;

		/// Weight multiplier.
		// float weightMultiplier;

		/// Factory size (0 to use the default "safe" value).
		unsigned int factorySize;

		/// Open node deque size (0 to use the default "safe" value).
		unsigned int openDequeSize;

		/// Closed node hash table precision.
		unsigned int hashPrecision;

		/// Weight (from 0 to 1) given to heuristic value (primarily used in A*).
		float heuristicWeight;

		/// Map width (set by createMap() from image passed).
		unsigned int width;
		/// Map height (set by createMap() from image passed).
		unsigned int height;

		double*		pMapData; // Values for the map data
	};


	//------------------------------------------------------------------------
	/** @name Construction/Destruction */
	//@{

	/// Constructor.
	Planner ();
	/// Destructor.
	virtual ~Planner ();

	//@}


	//------------------------------------------------------------------------
	/** @name Creating and Unloading */
	//@{

	/// Create a map to use for plotting.
	bool createMap (const Description &descrip);
	/// Unload any currently loaded map.
	void unloadMap ();

	/// Check if a map is loaded.
	bool isLoaded () const;

	//@}


	//------------------------------------------------------------------------
	/** @name Running */
	//@{

	/// Initialize the path-finding algorithm.
	virtual void initialize ();
	/// Run a single time slice of the path-finding algorithm.
	virtual bool run () = 0;
	/// Clean up nodes in use.
	virtual void cleanup ();

	//@}


	//------------------------------------------------------------------------
	/** @name Result Queries */
	//@{

	/// Check if the planner is done.
	bool isDone () const;

	/// Get the planner description.
	const Description &getDescription () const;

	/// Get the array of map values.
	const float *getMap () const;

	/// Get the goal node.
	const Node *getGoal () const;

	/// Get number of slices run.
	unsigned int getNumSlices () const;

	/// Get running time, in milliseconds.
	unsigned int getRunningTime () const;

	/// Get path cost.
	float getPathCost () const;

	/// Get the size of the path list.
	unsigned int getPathSize () const;
	/// Get the size of the open list.
	unsigned int getOpenSize () const;
	/// Get the size of the closed list.
	unsigned int getClosedSize () const;

protected:
	/// True if path planning is done, false if not.
	bool mt_bDone;

	/// Node factory.
	NodeFactory mt_nodeFactory;

	/// Description of map.
	Description mt_description;

	/// Map.
	float *mt_map;

	/// Number of slices run.
	int mt_numSlices;

	/// Time run so far, in milliseconds.
	unsigned int mt_runningTime;

	/// Open node list.
	FixedDeque<Node *> mt_open;
	/// Closed node hash table.
	NodeHashTable mt_closed;
	/// Final path (goal node).
	Node *mt_pGoal;
};


/// Check if a map is loaded.
/**
  @return  True if the planner currently has a map in memory, false if not.
*/
inline bool Planner::isLoaded () const
{
	return (mt_map != 0);
}


/// Run a single time slice of the path-finding algorithm.
/**
  @fn Planner::run()

  @return  True if the path-finding algorithm is done, false if not.

  @note  Before running the algorithm on recently loaded data, initialize()
         should be called first to set up the planner with whatever initial
         data it needs to run.

  @see initialize(), cleanup()
*/


/// Check if the planner is done.
/**
  @return  True if the planner is done checking for a path, false if not.

  @remarks  The planner may be done without having been able to find a path.
*/
inline bool Planner::isDone () const
{
	return mt_bDone;
}


/// Get the planner description.
/**
  This lets you get a hold of the description used to set up this planner.
  Some of the values may have been altered when creating the map, and this
  version of the description will represent those changes (if any).

  @return  Constant reference to the planner description.

  @note  The reference returned is a reference to the description structure
         stored within the planner, and as such would be invalid when the
         planner is deleted.
*/
inline const Planner::Description &Planner::getDescription () const
{
	return mt_description;
}


/// Get the array of map values.
/**
  This lets you access the actual array of map values used by the path finding
  data.

  @return  Constant pointer to the map data array.

  @note  The map returned is read-only.  The pointer given is to the actual
         map data, and as such would be invalid when a new map is loaded or
         the planner is deleted.
*/
inline const float *Planner::getMap () const
{
	return mt_map;
}


/// Get the goal node.
/**
  This returns the goal node, which should be linked to the previous nodes in
  the final path up to the start node using the Node::pParent variable.

  @return  Constant reference to the path list.

  @note  The pointer returned points to the actual node, and as such would be
         invalid when a new plan is started or the planner is deleted.
*/
inline const Node *Planner::getGoal () const
{
	return mt_pGoal;
}


/// Get number of slices run.
/**
  This gives you the number of slices (or the number of times run() has been
  callsed) used by this algorithm so far in planning a path.

  @return  Number of slices run.
*/
inline unsigned int Planner::getNumSlices () const
{
	return mt_numSlices;
}


/// Get running time, in milliseconds.
/**
  @return  Time spent searching for a path, in milliseconds.

  @remarks  The time returned is the time spent finding a path up to this
            point, or the total time if the algorithm is done searching for a
            path.  Time in between slices and time taken to fill in the path
            list itself are not taken into account.
*/
inline unsigned int Planner::getRunningTime () const
{
	return mt_runningTime;
}

#endif	// PLANNER_H
