/*
Copyright (c) 2007, Markus Trenkwalder

All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, 
this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution.

* Neither the name of the library's copyright owner nor the names of its 
contributors may be used to endorse or promote products derived from this 
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifdef TRENKI3D


#ifndef WIN32
#include <stdint.h>
#endif
#include "stdafx.h"

#include "16bitcolor.h"

// other includes
#include "trenki_vector_math.h"
#include "fixedpoint/fixed_class.h"

// requires software renderer source
#include "renderer/util.h"
#include "renderer/span.h"
#include "renderer/geometry_processor.h"
#include "renderer/rasterizer_subdivaffine.h"
using namespace swr;


#include "trenki3d.h"
#include "gfqueue.h"
#include "oscreen.h"

using namespace swr; // namespace for software renderer stuff
using namespace vmath; // vector math library
using namespace fixedpoint; // fixed point library

// some typedefs for easier use
#if 0 //def _DEBUG
	typedef float fixed16_t; // can be debugged, easier
	#define fixed_16_16_intvalue(a) (fixed_point<16>(a).intValue)
#else
	typedef fixed_point<16> fixed16_t;
	#define fixed_16_16_intvalue(a) ((a).intValue)
#endif

typedef vec2<fixed16_t> vec2x;
typedef vec3<fixed16_t> vec3x;
typedef vec4<fixed16_t> vec4x;

typedef mat2<fixed16_t> mat2x;
typedef mat3<fixed16_t> mat3x;
typedef mat4<fixed16_t> mat4x;


// vertex structure which will be used to store per vertex data which
// is passed to the vertex shader.
struct Trenki3DVertex {
	vec3x position;
	vec3x normal;
	fixed16_t   r,g,b;
	vec2x uv;
};

struct DiplayList
{
	Queue<Trenki3DVertex> vertices;
	Queue<unsigned int>   indices;
	int                   hasUV;
	int                   hasRGBA;
	int                   hasNormals;
};


// global depth buffer we will use.
// global drawing surface.
struct TRENKI_GL
{
	TRENKI_GL() :g(&r),depth_buffer(0),color_buffer(0),screen_w(0),
				m_pTexels(0), m_TexW(0), m_TexH(0),
				m_MatStackDepth(0), m_Lighting(0)
			{}

	unsigned short* depth_buffer, *color_buffer;
	unsigned int    screen_w;
	// create a rasterizer class that will be used to rasterzie primitives.
	RasterizerSubdivAffine r;
	// create a geometry processor class used to feed vertex data.
	GeometryProcessor g;

	mat4x	m_ProjectionMatrix;
	mat4x	m_MatStack[16];
	mat4x	m_FinalMatrix; // just before drawing
	vec3x	m_LightDir;
	int		m_Lighting;
	int		m_zTest;
	int		m_MatStackDepth;
	int		m_iCurDisplayList; // -1 for none
	Queue<DiplayList > m_DisplayLists;
	
	unsigned short* m_pTexels;
	unsigned short  m_CookieColor;
	bool			m_CookieCheck;
	int m_TexW, m_TexH;
	
} gTgl;



// this is the vertex shader which is executed for each individual vertex that
// needs to be processed.
template <const int Ttex, const int Tcol, const int Tlight>
struct MyVertexShader {
	// this specifies that this shader is only going to use 1 vertex attribute 
	// array. there can be used up to GeometryProcessor::MAX_ATTRIBUTES arrays.
	// this needs to be static but does not necessarily have to be const.
	static const unsigned int attribute_count=3; // = Ttex*2 + Tcol*3 + Tlight;
	static const unsigned int varying_count;

	// this static function is called for each vertex to be processed.
	// "in" is an array of void* pointers with the location of the individial 
	// vertex attributes. the "out" structure has to be written to.
	static void shade(const GeometryProcessor::VertexInput in, GeometryProcessor::VertexOutput &out)
	{
		// some constants used
		static const fixed16_t half = 0.5f;
		static const fixed16_t one = 1.0f;

		// cast the first attribute array to the input vertex type.
		Trenki3DVertex &v = *(Trenki3DVertex*)(in[0]);

		// transform the vertex by the transformation matrix
		vec4x t = gTgl.m_FinalMatrix * vec4x(v.position, one);

		// x, y, z and w are the components that must be written by the vertex
		// shader. they all have to be specified in 16.16 fixed point format.
		out.x = fixed_16_16_intvalue(t.x);
		out.y = fixed_16_16_intvalue(t.y);
		out.z = fixed_16_16_intvalue(t.z);
		out.w = fixed_16_16_intvalue(t.w);

		// write the varying which will be interpolated across the triangle.
		int ivar = 0;
		if(Ttex)
		{
			out.varyings[ivar++] = fixed_16_16_intvalue(v.uv[0]);
			out.varyings[ivar++] = fixed_16_16_intvalue(v.uv[1]);
		}
		if(Tcol)
		{
			out.varyings[ivar++] = fixed_16_16_intvalue(v.r);
			out.varyings[ivar++] = fixed_16_16_intvalue(v.g);
			out.varyings[ivar++] = fixed_16_16_intvalue(v.b);
		}
		if(Tlight)
		{
			// calculate the lighting. doing it this way we will also get some nice
			// shading on the back side of the model.
			fixed16_t lighting = dot(v.normal, gTgl.m_LightDir) * half + half;
			out.varyings[ivar++] = 31 * fixed_16_16_intvalue(lighting);
		}
	}
};

// Freaking GCC needs these
#define VARCNT(t,c,l) template<> const unsigned int MyVertexShader<t,c,l>::varying_count = t*2 + c*3 + l;
VARCNT(1,1,1);
VARCNT(1,1,0);
VARCNT(1,0,1);
VARCNT(1,0,0);
VARCNT(0,1,1);
VARCNT(0,1,0);
VARCNT(0,0,1);
VARCNT(0,0,0);
#undef VARCNT


void (*gTrenki3DBlendFunc)(unsigned short* pdest, unsigned short src);


// this is the fragment shader.
template <const int Ttex, const int Tcol, const int Tlight, int TzBuffer, const int AlphaShite>
struct MyFragmentShader : public GenericSpanDrawer<MyFragmentShader<Ttex, Tcol, Tlight, TzBuffer, AlphaShite> >
{
	// varying_count = 1 tells the rasterizer that it only needs to interpolate
	// one varying value which will be the lighting value computed per vertex.
	static const unsigned int varying_count; // = Ttex*2 + Tcol*3 + Tlight;

	// we also want to interpolate the depth value because we do depth testing.
	static const int interpolate_z = TzBuffer;

	// per triangle callback. this could for instance be used to select the 
	// mipmap level of detail. we don't need it but it still needs to be defined
	// for everything to work.
	static void begin_triangle(
		const IRasterizer::Vertex& /*v1*/, const IRasterizer::Vertex& /*v2*/, const IRasterizer::Vertex& /*v3*/, int /*area2*/
		)	{ }

	// this static function will be called for every pixel that is to be 
	// rasterized. x and y are the screen coordinates and fd holds the data for
	// the fragment that should be drawn.
	static void single_fragment(int x, int y, const IRasterizer::FragmentData &fd)
	{
		if(TzBuffer)
		{
			// get the depth buffer pixel from x, y parameters.
			short *db = (short*)gTgl.depth_buffer + x + y * Screen.db_width; // #NEED z buffer width != sprite width?
			
			// do the depth test (note: we have a 16bit depth buffer, therefore the shift).
			short depth = fd.z >> 16;
			if (*db < depth)
				return;
			*db = depth;
		}
		
		// get the color buffer pixel.
		unsigned short *color_buffer = gTgl.color_buffer + x + y * gTgl.screen_w;
		//VALID_SCREEN(color_buffer);

		unsigned int r,g,b;
		int ivar = 0;
		if(Ttex)
		{
			// limit to 0..32767
			unsigned int tu = (unsigned)fd.varyings[ivar++] & 0xffff;
			unsigned int tv = (unsigned)fd.varyings[ivar++] & 0xffff;
			tu*=(gTgl.m_TexW); tu>>=16; // must be this way - don't subtract one or you get bad mapping!
			tv*=(gTgl.m_TexH); tv>>=16;
			
			// tv = gTgl.m_TexH - tv;
			unsigned short* pTex = gTgl.m_pTexels + tu + (gTgl.m_TexH - tv-1) * gTgl.m_TexW;
			
			if(*pTex == gTgl.m_CookieColor && gTgl.m_CookieCheck) return;

			// shortcut if no light active
			if(!Tcol)
			{
				*color_buffer = *pTex;
				return;
			}
			r = GETR(*pTex);
			g = GETG(*pTex);
			b = GETB(*pTex);
			if(Tcol)
			{
				r = (r*fd.varyings[ivar++]) >> 16;
				g = (g*fd.varyings[ivar++]) >> 16;
				b = (b*fd.varyings[ivar++]) >> 16;
			}
		}
		else
		{
			if(Tcol)
			{
				r = (255*fd.varyings[ivar++]) >> 16;
				g = (255*fd.varyings[ivar++]) >> 16;
				b = (255*fd.varyings[ivar++]) >> 16;
			}
			else
			{
				r=g=b= 255;
			}
		}
		
		
		if(Tlight)
		{
			// make sure the interpolated value lies in the correct range.
			int s = std::min(std::max(fd.varyings[ivar++] >> 16, 0), 255);
			r = (s*r)>>8;
			g = (s*g)>>8;
			b = (s*b)>>8;
		}

		// write the color to the buffer.
		if(AlphaShite)
			gTrenki3DBlendFunc(color_buffer, PackColor(r,g,b));
		else
			*color_buffer = PackColor(r,g,b);
	}
};

// make the linker happy (MSVC doesn't complain if this
// is missing while GCC does).
#define VARCNT(t,c,l,z,a) template<> const unsigned int MyFragmentShader<t,c,l,z,a>::varying_count = t*2 + c*3 + l;
VARCNT(1,1,1,1,0);
VARCNT(1,1,1,0,0);
VARCNT(1,1,0,1,0);
VARCNT(1,1,0,0,0);
VARCNT(1,0,1,1,0);
VARCNT(1,0,1,0,0);
VARCNT(1,0,0,1,0);
VARCNT(1,0,0,0,0);
VARCNT(0,1,1,1,0);
VARCNT(0,1,1,0,0);
VARCNT(0,1,0,1,0);
VARCNT(0,1,0,0,0);
VARCNT(0,0,1,1,0);
VARCNT(0,0,1,0,0);
VARCNT(0,0,0,1,0);
VARCNT(0,0,0,0,0);

VARCNT(1,1,1,1,1);
VARCNT(1,1,1,0,1);
VARCNT(1,1,0,1,1);
VARCNT(1,1,0,0,1);
VARCNT(1,0,1,1,1);
VARCNT(1,0,1,0,1);
VARCNT(1,0,0,1,1);
VARCNT(1,0,0,0,1);
VARCNT(0,1,1,1,1);
VARCNT(0,1,1,0,1);
VARCNT(0,1,0,1,1);
VARCNT(0,1,0,0,1);
VARCNT(0,0,1,1,1);
VARCNT(0,0,1,0,1);
VARCNT(0,0,0,1,1);
VARCNT(0,0,0,0,1);

#undef VARCNT


// conversion function used to convert the float vectors 
// loaded from the .obj file to fixed point.
inline vec3x convert(const vec3<float> & rhs)
{
	return vec3x(rhs.x, rhs.y, rhs.z);
}



// glWRAPPERS
void tglInit(unsigned short* pPix, unsigned short* pZBuf, unsigned int width)
{
	gTgl.m_CookieColor = PackColor(255,0,128);
	gTgl.m_CookieCheck = true;
	gTgl.depth_buffer = pZBuf;
	gTgl.color_buffer = pPix;
	gTgl.screen_w = width;
	gTgl.m_zTest=0;
	gTgl.m_iCurDisplayList=-1;
	gTgl.m_DisplayLists.Clear();
	
	// disable perspective correction for faster rendering.
	// in this demo the difference won't be noticeable since the triangles 
	// are all small enough and we only use vertex colors and no textures.
	gTgl.r.perspective_correction(false);
	gTrenki3DBlendFunc = NULL;
}

void tglSetCookieColor(unsigned short packed_color, bool do_check)
{
	gTgl.m_CookieColor = packed_color;
	gTgl.m_CookieCheck = do_check;
}


void tglSetBuffers(unsigned short* pPix, unsigned short* pZBuf, int buffer_w)
{
	gTgl.depth_buffer = pZBuf;
	gTgl.color_buffer = pPix;
	gTgl.screen_w = buffer_w;
}

void tglEnableDepthTest(int ok)
{
	gTgl.m_zTest = ok ? 1:0;

	// with a z-buffer, enable depth correction
	gTgl.r.perspective_threshold(16,16);
	gTgl.r.perspective_correction(ok ? true:false);
}

int tglIsEnabledDepthTest()
{
	return gTgl.m_zTest;
}


void tglViewport(int x, int y, int w, int h)
{
	// it is necessary to set the viewport.
	gTgl.g.viewport(x,y, w, h);

	// it is also necessary to set the clipping rectangle.
	gTgl.r.clip_rect(x,y, w, h);
	if(y+h>480)
	{
		int a=5;
	}
}

#define GL_FRONT	0x0404
#define GL_BACK		0x0405

void tglCullFace(int mode)
{
	// set the cull mode (cw is already the default mode).
	if(mode == GL_FRONT)
		gTgl.g.cull_mode(GeometryProcessor::CULL_CW);
	else if(mode== GL_BACK)
		gTgl.g.cull_mode(GeometryProcessor::CULL_CCW);
	else
		gTgl.g.cull_mode(GeometryProcessor::CULL_NONE);
}



void tglLookAt(float eyex, float eyey, float eyez, float centerx, float centery, float centerz, float upx, float upy, float upz)
{
	vec3x eye = vec3x(fixed16_t(eyex), fixed16_t(eyey), fixed16_t(eyez));
	vec3x center = vec3x(fixed16_t(centerx), fixed16_t(centery), fixed16_t(centerz));
	vec3x up = vec3x(fixed16_t(upx), fixed16_t(upy), fixed16_t(upz));
	gTgl.m_MatStack[gTgl.m_MatStackDepth] = lookat_matrix(eye, center, up);
}


void tglOrtho(int l, int r, int b, int t, float n, float f)
{
	gTgl.m_ProjectionMatrix=ortho_matrix<fixed16_t>(fixed16_t(l),fixed16_t(r),fixed16_t(b),fixed16_t(t),fixed16_t(n),fixed16_t(f));
}

void tglFrustum(float l, float r, float b, float t, float n, float f)
{
	gTgl.m_ProjectionMatrix=frustum_matrix<fixed16_t>(fixed16_t(l), fixed16_t(r),fixed16_t(b), fixed16_t(t),fixed16_t(n), fixed16_t(f));
}

void tgluPerspective(float fov, float aspect, float zNear, float zFar)
{
	gTgl.m_ProjectionMatrix=perspective_matrix<fixed16_t>(fixed16_t(fov), fixed16_t(aspect),fixed16_t(zNear), fixed16_t(zFar));
}

void tglLoadIdentity()
{
	gTgl.m_MatStack[gTgl.m_MatStackDepth] = identity4<fixed16_t>();
}


void tglPushMatrix()
{
	gTgl.m_MatStack[gTgl.m_MatStackDepth+1] = gTgl.m_MatStack[gTgl.m_MatStackDepth];
	gTgl.m_MatStackDepth++;
}

void tglPopMatrix()
{
	if(gTgl.m_MatStackDepth>0)
		--gTgl.m_MatStackDepth;
}

void tglMultMatrixf(float* mat)
{
fixed16_t m0 = mat[0 ];
fixed16_t m1 = mat[1 ];
fixed16_t m2 = mat[2 ];
fixed16_t m3 = mat[3 ];
fixed16_t m4 = mat[4 ];
fixed16_t m5 = mat[5 ];
fixed16_t m6 = mat[6 ];
fixed16_t m7 = mat[7 ];
fixed16_t m8 = mat[8 ];
fixed16_t m9 = mat[9 ];
fixed16_t m10= mat[10];
fixed16_t m11= mat[11];
fixed16_t m12= mat[12];
fixed16_t m13= mat[13];
fixed16_t m14= mat[14];
fixed16_t m15= mat[15];
/*
mat4x xmat (m0,m1, m2, m3,
				m4,m5, m6, m7,
				m8,m9, m10,m11,
				m12,m13,m14,m15);
*/
mat4x xmat (	m0 ,m4 , m8 , m12,
				m1 ,m5 , m9 , m13,
				m2 ,m6 , m10, m14,
				m3 ,m7 , m11, m15);
	gTgl.m_MatStack[gTgl.m_MatStackDepth] *= xmat;
}

void tglBindTexture(unsigned short* pTex, int width, int height)
{
	gTgl.m_TexW = width;
	gTgl.m_TexH = height;
	gTgl.m_pTexels = pTex;
}


void tglSetAlphaFunc(	void (*BlendFunc)(unsigned short* pdest, unsigned short src) )
{
	gTrenki3DBlendFunc = BlendFunc;
}


void tglInternalDraw_alpha(DiplayList& dlist)
{
	#define HasAlphaBlending 1

	// set the vertex and fragment shaders.
	if(dlist.hasUV && gTgl.m_pTexels)
	{
		// < Texture, Colors, Light >
		if(dlist.hasRGBA)
		{
			if(dlist.hasNormals && gTgl.m_Lighting)
			{
				gTgl.g.vertex_shader<MyVertexShader    <1,1,1> >();
				if(gTgl.m_zTest)
					gTgl.r.fragment_shader<MyFragmentShader<1,1,1,1,HasAlphaBlending> >();
				else
					gTgl.r.fragment_shader<MyFragmentShader<1,1,1,0,HasAlphaBlending> >();
			}
			else
			{
				gTgl.g.vertex_shader<MyVertexShader    <1,1,0> >();
				if(gTgl.m_zTest)
					gTgl.r.fragment_shader<MyFragmentShader<1,1,0,1,HasAlphaBlending> >();
				else
					gTgl.r.fragment_shader<MyFragmentShader<1,1,0,0,HasAlphaBlending> >();
			}
		}
		else
		{
			if(dlist.hasNormals && gTgl.m_Lighting)
			{
				gTgl.g.vertex_shader<MyVertexShader    <1,0,1> >();
				if(gTgl.m_zTest)
					gTgl.r.fragment_shader<MyFragmentShader<1,0,1,1,HasAlphaBlending> >();
				else
					gTgl.r.fragment_shader<MyFragmentShader<1,0,1,0,HasAlphaBlending> >();
			}
			else
			{
				gTgl.g.vertex_shader<MyVertexShader    <1,0,0> >();
				if(gTgl.m_zTest)
					gTgl.r.fragment_shader<MyFragmentShader<1,0,0,1,HasAlphaBlending> >();
				else
					gTgl.r.fragment_shader<MyFragmentShader<1,0,0,0,HasAlphaBlending> >();
			}
		}	
	}
	else
	{
		if(dlist.hasRGBA)
		{
			if(dlist.hasNormals && gTgl.m_Lighting)
			{
				gTgl.g.vertex_shader<MyVertexShader    <0,1,1> >();
				if(gTgl.m_zTest)
					gTgl.r.fragment_shader<MyFragmentShader<0,1,1,1,HasAlphaBlending> >();
				else
					gTgl.r.fragment_shader<MyFragmentShader<0,1,1,0,HasAlphaBlending> >();
			}
			else
			{
				gTgl.g.vertex_shader<MyVertexShader    <0,1,0> >();
				if(gTgl.m_zTest)
					gTgl.r.fragment_shader<MyFragmentShader<0,1,0,1,HasAlphaBlending> >();
				else
					gTgl.r.fragment_shader<MyFragmentShader<0,1,0,0,HasAlphaBlending> >();
			}
		}
		else
		{
			if(dlist.hasNormals && gTgl.m_Lighting)
			{
				gTgl.g.vertex_shader<MyVertexShader    <0,0,1> >();
				if(gTgl.m_zTest)
					gTgl.r.fragment_shader<MyFragmentShader<0,0,1,1,HasAlphaBlending> >();
				else
					gTgl.r.fragment_shader<MyFragmentShader<0,0,1,0,HasAlphaBlending> >();
			}
			else
			{
				gTgl.g.vertex_shader<MyVertexShader    <0,0,0> >();
				if(gTgl.m_zTest)
					gTgl.r.fragment_shader<MyFragmentShader<0,0,0,1,HasAlphaBlending> >();
				else
					gTgl.r.fragment_shader<MyFragmentShader<0,0,0,0,HasAlphaBlending> >();
			}
		}
	}
	#undef HasAlphaBlending
}

void tglInternalDraw_no_alpha(DiplayList& dlist)
{
	#define HasAlphaBlending 0

	// set the vertex and fragment shaders.
	if(dlist.hasUV && gTgl.m_pTexels)
	{
		// < Texture, Colors, Light >
		if(dlist.hasRGBA)
		{
			if(dlist.hasNormals && gTgl.m_Lighting)
			{
				gTgl.g.vertex_shader<MyVertexShader    <1,1,1> >();
				if(gTgl.m_zTest)
					gTgl.r.fragment_shader<MyFragmentShader<1,1,1,1,HasAlphaBlending> >();
				else
					gTgl.r.fragment_shader<MyFragmentShader<1,1,1,0,HasAlphaBlending> >();
			}
			else
			{
				gTgl.g.vertex_shader<MyVertexShader    <1,1,0> >();
				if(gTgl.m_zTest)
					gTgl.r.fragment_shader<MyFragmentShader<1,1,0,1,HasAlphaBlending> >();
				else
					gTgl.r.fragment_shader<MyFragmentShader<1,1,0,0,HasAlphaBlending> >();
			}
		}
		else
		{
			if(dlist.hasNormals && gTgl.m_Lighting)
			{
				gTgl.g.vertex_shader<MyVertexShader    <1,0,1> >();
				if(gTgl.m_zTest)
					gTgl.r.fragment_shader<MyFragmentShader<1,0,1,1,HasAlphaBlending> >();
				else
					gTgl.r.fragment_shader<MyFragmentShader<1,0,1,0,HasAlphaBlending> >();
			}
			else
			{
				gTgl.g.vertex_shader<MyVertexShader    <1,0,0> >();
				if(gTgl.m_zTest)
					gTgl.r.fragment_shader<MyFragmentShader<1,0,0,1,HasAlphaBlending> >();
				else
					gTgl.r.fragment_shader<MyFragmentShader<1,0,0,0,HasAlphaBlending> >();
			}
		}	
	}
	else
	{
		if(dlist.hasRGBA)
		{
			if(dlist.hasNormals && gTgl.m_Lighting)
			{
				gTgl.g.vertex_shader<MyVertexShader    <0,1,1> >();
				if(gTgl.m_zTest)
					gTgl.r.fragment_shader<MyFragmentShader<0,1,1,1,HasAlphaBlending> >();
				else
					gTgl.r.fragment_shader<MyFragmentShader<0,1,1,0,HasAlphaBlending> >();
			}
			else
			{
				gTgl.g.vertex_shader<MyVertexShader    <0,1,0> >();
				if(gTgl.m_zTest)
					gTgl.r.fragment_shader<MyFragmentShader<0,1,0,1,HasAlphaBlending> >();
				else
					gTgl.r.fragment_shader<MyFragmentShader<0,1,0,0,HasAlphaBlending> >();
			}
		}
		else
		{
			if(dlist.hasNormals && gTgl.m_Lighting)
			{
				gTgl.g.vertex_shader<MyVertexShader    <0,0,1> >();
				if(gTgl.m_zTest)
					gTgl.r.fragment_shader<MyFragmentShader<0,0,1,1,HasAlphaBlending> >();
				else
					gTgl.r.fragment_shader<MyFragmentShader<0,0,1,0,HasAlphaBlending> >();
			}
			else
			{
				gTgl.g.vertex_shader<MyVertexShader    <0,0,0> >();
				if(gTgl.m_zTest)
					gTgl.r.fragment_shader<MyFragmentShader<0,0,0,1,HasAlphaBlending> >();
				else
					gTgl.r.fragment_shader<MyFragmentShader<0,0,0,0,HasAlphaBlending> >();
			}
		}
	}
	#undef HasAlphaBlending
}

void tglPerformRender(DiplayList& dlist)
{
	if(!dlist.indices.Size()) return;

	if(gTrenki3DBlendFunc)
		tglInternalDraw_alpha(dlist);
	else
		tglInternalDraw_no_alpha(dlist);
	

	// Specify where out data lies in memory
	gTgl.g.vertex_attrib_pointer(0, sizeof(Trenki3DVertex), &dlist.vertices[0]);
	
	gTgl.m_FinalMatrix = gTgl.m_ProjectionMatrix * gTgl.m_MatStack[gTgl.m_MatStackDepth];
	
	unsigned int* ptI = &dlist.indices[0];
	int count = (int)dlist.indices.Size();
	while(count>2)
	{
		// draw the triangle
		gTgl.g.draw_triangles(3, ptI); // #NEED can't draw all at once!?
		ptI+=3;
		count-=3;
	}
}



void tglDrawArraysXYZ_UV_RGBA(const int type, const int count, const float* pData, const int hasZ, const int hasNormal, const int UseUV, const int UseRGBA, const int hasRGBA)
{
static DiplayList dlist;
	dlist.indices.AllocAtLeast (count*3);
	dlist.indices.SetSize(0);
	dlist.vertices.SetSize(count);
	dlist.hasNormals = hasNormal;
	dlist.hasRGBA    = UseRGBA;
	dlist.hasUV      = UseUV;

int A=0,B=0,C=0;
int even = 0;
int ntri=0;
	#define PUSHTRI() {dlist.indices.Push(A); dlist.indices.Push(B); dlist.indices.Push(C);}
	#define PUSHTRI_INV() {dlist.indices.Push(C); dlist.indices.Push(B); dlist.indices.Push(A);}

	for(int i=0; i<count; ++i)
	{
		switch(type)
		{
		case GL_TRIANGLE_FAN:
			B = C;
			C = i;
			if(i>1) PUSHTRI();
			break;
		case GL_TRIANGLES:
			A=B; B=C; C=i;
			if((i%3) == 2) PUSHTRI();
			break;
		case GL_TRIANGLE_STRIP:
			++ntri;
			A=B; B=C; C=i;
			if(ntri==3)
			{
				even=even ? 0:1;
				ntri=2;
				if(even)
				{
					PUSHTRI();
				}
				else
				{
					PUSHTRI_INV();
				}
				// strip has every even triangle reversed
				// int swap=B; B=C; C=swap;
			}
			break;
		}
		
		Trenki3DVertex* pC = &dlist.vertices[i];
		
		pC->position.x = fixed16_t(*pData++);
		pC->position.y = fixed16_t(*pData++);
		if(hasZ)
		{
			pC->position.z = fixed16_t(*pData++);
		}
		else
		{
			pC->position.z = fixed16_t(0);
		}
		
		if(hasNormal)
		{
			pC->normal.x = fixed16_t(*pData++);
			pC->normal.y = fixed16_t(*pData++);
			if(hasZ)
			{
				pC->normal.z = fixed16_t(*pData++);
			}
		}
		
		if(UseUV)
		{
			pC->uv.x = fixed16_t(*pData++);
			pC->uv.y = fixed16_t(*pData++);
		}
		else
		{
			pData+=2;
		}

		// TRENKI_COL_PRT
		if(UseRGBA)
		{
			const unsigned char* pColors = ((const unsigned char*)pData++);
			const float ffac = 1.0f / 255.0f;
			pC->r = fixed16_t(ffac * (float)pColors[0]);
			pC->g = fixed16_t(ffac * (float)pColors[1]);
			pC->b = fixed16_t(ffac * (float)pColors[2]);
			/*float alpha =ffac * pColors[3]; */
		}
		else
		{
			pC->r = pC->g = pC->b = fixed16_t(1.0f);
			if(hasRGBA)
				pData++; // not _using_ RGBA, but stored!
		}
		++pC;
	}


	if(gTgl.m_iCurDisplayList < 0)
	{
		tglPerformRender(dlist);
	}
	else
	{
		gTgl.m_DisplayLists[gTgl.m_iCurDisplayList] = dlist;
	}
}


int tglGenLists(int size)
{
	int pos = gTgl.m_DisplayLists.Size();
	gTgl.m_DisplayLists.SetSize(pos + size);
	return pos;
}


void tglNewList(int iList)
{
	gTgl.m_DisplayLists[iList].vertices.Clear(false);
	gTgl.m_DisplayLists[iList].indices.Clear(false);
	gTgl.m_iCurDisplayList = iList;
}


void tglEndList()
{
	gTgl.m_iCurDisplayList = -1;
}

void tglCallLists(int iFirst, int count)
{
	for(int i=0; i<count; ++i)
	{
		tglPerformRender(gTgl.m_DisplayLists[i+iFirst]);
	}
}


// GP2X - upper 32MB memory
void * UpperMalloc(size_t size);
void UpperFree(void* mem);
int GetUpperSize(void* mem);
void InitMemPool();
void DestroyMemPool();

#if defined(GP2X) // || defined(_DEBUG)
	#define qmalloc UpperMalloc
	#define qfree   UpperFree

// Using GP2X upper memory

#include <stdlib.h>
#if defined(GP2X) // only GP2X, WIZ uses new/delete
	#include <stdio.h>
	#include <fcntl.h>
	#include <sys/mman.h>
	#include <linux/fb.h>
	#include <unistd.h>
	#include <stropts.h>
	volatile void *UpperMem=NULL;
#else
	void *UpperMem=NULL;
#endif


volatile int Uppermemfd;
#define BLOCKSIZE 1024
int TakenSize[0x2000000 / BLOCKSIZE];
#define SetTaken(Start, Size) TakenSize[(Start - 0x2000000) / BLOCKSIZE] = (Size - 1) / BLOCKSIZE + 1


void * UpperMalloc(size_t size) {
  int i = 0;
ReDo:
  for (; TakenSize[i]; i += TakenSize[i]);
  if (i >= 0x2000000 / BLOCKSIZE) {
    printf("UpperMalloc out of mem!");
    return NULL;
  }
  int BSize = (size - 1) / BLOCKSIZE + 1;
  for(int j = 1; j < BSize; j++) {
    if (TakenSize[i + j]) {
      i += j;
      goto ReDo;
    }
  }
  
  TakenSize[i] = BSize;
  void* mem = ((char*)UpperMem) + i * BLOCKSIZE;
  return mem;
}

//Releases UpperMalloced memory
void UpperFree(void* mem) {
  int i = (((int)mem) - ((int)UpperMem));
  if (i < 0 || i >= 0x2000000) {
    fprintf(stderr, "UpperFree of not UpperMalloced mem: %p\n", mem);
  } else {
    if (i % BLOCKSIZE)
      fprintf(stderr, "delete error: %p\n", mem);
    TakenSize[i / BLOCKSIZE] = 0;
  }
}

//Returns the size of a UpperMalloced block.
int GetUpperSize(void* mem) {
  int i = (((int)mem) - ((int)UpperMem));
  if (i < 0 || i >= 0x2000000) {
    fprintf(stderr, "GetUpperSize of not UpperMalloced mem: %p\n", mem);
    return -1;
  }
  return TakenSize[i / BLOCKSIZE] * BLOCKSIZE;
}

//Returns hardware address for buffer.
#ifdef GP2X
	unsigned int GetUpperRealAddress(void* pVirtual)
	{
	#ifdef _DEBUG
		if ((unsigned int)pVirtual < 0 || (unsigned int)pVirtual >= 0x2000000)
			*(int*)0=42;
	#endif
		return ((unsigned int)pVirtual-(unsigned int)UpperMem+0x2000000);
	}
#endif

void InitMemPool()
{
if(UpperMem) return;

#ifdef GP2X
	FILE* pF = fopen("/etc/open2x", "rb");
	if(pF)
	{
		fclose(pF);
		printf("On Open2X modprobing mmuhack.o\n");
		system("/sbin/modprobe mmuhack");
	}
	else
	{
		//Try to apply MMU hack.
		int mmufd = open("/dev/mmuhack", O_RDWR);
		if(mmufd < 0)
		{
			system("/sbin/insmod mmuhack.o");
			mmufd = open("/dev/mmuhack", O_RDWR);
		}
		if(mmufd < 0)
		{
			printf("MMU hack failed\n");
			// exit(42); -> just slower
		}
		else
		{
			close(mmufd);
		}
	}

  Uppermemfd = open("/dev/mem", O_RDWR);
  UpperMem = mmap(0, 0x2000000, PROT_READ | PROT_WRITE, MAP_SHARED, Uppermemfd, 0x2000000);
  printf("Got /dev/mem at %p\n", Uppermemfd);
#else
  UpperMem = malloc(0x2000000);
#endif

#ifdef GPHWIZ
	void wiz_mmuhack(int mem_fd);
	int mem_fd = open("/dev/mem");
	wiz_mmuhack(mem_fd);
	close(mem_fd);
#endif

  memset(TakenSize, 0, sizeof(TakenSize));

  SetTaken(0x3000000, 0x80000); // Video decoder (you could overwrite this, but if you
                                // don't need the memory then be nice and don't)
  SetTaken(0x3101000, 153600);  // Primary frame buffer
  SetTaken(0x3381000, 153600);  // Secondary frame buffer (if you don't use it, uncomment)
  SetTaken(0x3600000, 0x8000);  // Sound buffer
}

void DestroyMemPool()
{
if(!UpperMem) return;
#ifdef GP2X
  close (Uppermemfd);
#else
  free(UpperMem);
#endif
  UpperMem = NULL;
}
#else // PocketPC or something
	#define qmalloc(a) ((void*)new unsigned char[(a)])
	#define qfree(a)   delete [](unsigned char*)(a);
	void InitMemPool()    {}
	void DestroyMemPool() {}
#endif // GP2X | DEBUG


void gl_free(void *p)
{
    qfree(p);
}

void *gl_malloc(int size)
{
    return qmalloc(size);
}

void *gl_zalloc(int size)
{
void* pm = qmalloc(size);
	if(pm) memset(pm,0,size);
	return pm;
}



#endif // TRENKI3D
