// GFini - ini file handling for x-platformers

#ifndef _GF_INI_H_INC_
#define _GF_INI_H_INC_

#include "gffilesystem.h"
#include "gfqueue.h"


#ifdef WANT_SHOEBOX
	#include "gfshoebox.h"
	extern GFShoeBox *__g_ShoeBox_Data;
#endif

#ifndef stricmp
#define stricmp _stricmp
#endif

namespace __GLBASIC__ {
BOOL glb_prepare_reading_file(DGStr& filename);
};

// --------------------------------------------------------------- //
// A robust, fast class for ini file handling
// --------------------------------------------------------------- //
class GFini
{
	struct INIKEY
	{
		DGStr name, val;
	};
	struct SECTION
	{
		DGStr			name;
		Queue<INIKEY>	keys;
	};
	Queue<SECTION> m_sections;
	DGStr m_filename;

public:
	bool m_changed;
	GFini()  {Clear();}
	~GFini() {Write();}
	
	void Clear()
	{
		m_changed=false;
		m_filename = CGStr("");
		m_sections.Clear();
	}

	// --------------------------------------------------------------- //
	// Read the file
	// --------------------------------------------------------------- //
	void Read(const DGStr& fn)
	{
		Clear();
		DGStr filename(fn);
		if(!__GLBASIC__::glb_prepare_reading_file(filename))
			filename = fn;

		m_sections.SetAllocSteps(16,8);

		char buf[1024], *pCur=NULL;
		GFFileSystem fs;
		m_filename=filename;

		if(!fs.OpenIn(filename.c_str()))
		{
			return;
		}

		INIKEY key;
		SECTION section;
		section.keys.SetAllocSteps(32, 32);
		
		while(fs.ReadLine(buf, 1024) )
		{
			#define SKIPWS while(*pCur==' '||*pCur=='\t'||*pCur=='\r'||*pCur=='\n')++pCur;
			#define SKIPNW while(*pCur!=' '&&*pCur!='\t'&&*pCur!='\r'&&*pCur!='\n'&&*pCur!='\0'&&*pCur!='=')++pCur;

			pCur = buf;
			SKIPWS;
			char* pName = pCur;
			if(*pName == ';') continue; // comment!

			// new section
			if(*pName == '[')
			{
				if (section.keys.Size())
				{
					m_sections.Push(section);
					section.keys.Clear(false);
				}
				++pCur;
				SKIPWS;
				char* pSec = pCur;
				SKIPNW;
				--pCur;
				// [ section ] or [section] ?
				if (*pCur==']') *pCur='\0';
					else pCur[1] = '\0';
				if(*pSec=='\0') continue; // uh-oh!
				section.name.assignlowmem(pSec);
				continue;
			}
			
			// is a value?
			SKIPNW;
			if(*pCur=='\0') continue; // some bug
			bool has_eq = false;
			if(*pCur=='=') has_eq=true;
			*pCur = '\0';
			key.name.assignlowmem(pName);
			++pCur;
			if(!has_eq)
			{
				SKIPWS;
				if(*pCur!='=') continue; // wrong line
				++pCur;			
			}
			SKIPWS;
			if (*pCur=='\0') continue; // empty value
			key.val.assignlowmem(pCur);
			key.val.trimright();
			section.keys.Push(key);
			#undef SKIPWS
			#undef SKIPNW
		}
		if (section.keys.Size())
		{
			m_sections.Push(section);
			section.keys.Clear(false);
		}
		fs.Close();
	}
	
	// --------------------------------------------------------------- //
	// Write the file if changed
	// --------------------------------------------------------------- //
	void Write()
	{
		if(!m_changed || m_filename.len()==0) return;

		GFFileSystem fs;
		fs.OpenOut(m_filename.c_str());
		for(int i=0; i<(int)m_sections.Size(); ++i)
		{
			fs.WriteLine((CGStr("[")+m_sections[i].name+CGStr("]")).c_str());
			for(int j=0; j<(int)m_sections[i].keys.Size(); ++j)
			{
				fs.WriteLine((m_sections[i].keys[j].name+CGStr("=")+m_sections[i].keys[j].val).c_str());
			}
		}
		fs.Close();
	}
	
	// --------------------------------------------------------------- //
	// find index of section
	// --------------------------------------------------------------- //
	int FindSection(const DGStr& section)
	{
		for(int i=0; i<(int)m_sections.Size(); ++i)
		{
			if(stricmp(section.c_str(), m_sections[i].name.c_str())==0)
				return i;
		}
	return -1;
	}

	// --------------------------------------------------------------- //
	// Find index of value in section
	// --------------------------------------------------------------- //
	int FindVal(int section, const DGStr& value)
	{
		for(int i=0; i<(int)m_sections[section].keys.Size(); ++i)
		{
			if(stricmp(value.c_str(), m_sections[section].keys[i].name.c_str())==0)
				return i;
		}
	return -1;
	}
	
	// --------------------------------------------------------------- //
	// Remove a complete section
	// --------------------------------------------------------------- //
	void RemoveSection(const DGStr& section)
	{
		int sec = FindSection(section);
		if(sec>=0)
		{
			m_sections.Remove(sec, true);
			m_changed=true;
		}
	}

	// --------------------------------------------------------------- //
	// Get a Value. NO_DATA if not there
	// --------------------------------------------------------------- //
	DGStr GetVal(const DGStr& section, const DGStr& key, const DGStr& nodata)
	{
		int sec = FindSection(section);
		if(sec<0) return nodata;

		int val= FindVal(sec, key);
		if(val<0)
		{
			return nodata;
		}
		return m_sections[sec].keys[val].val;
	}
	
	// --------------------------------------------------------------- //
	// Set a value. If val="", the val get's removed
	// --------------------------------------------------------------- //
	void SetVal(const DGStr& section, const DGStr& key, const DGStr& value)
	{
		int sec = FindSection(section);
		// insert new section?
		if(sec<0)
		{
			SECTION sect;
			sect.name = section;
			m_sections.Push(sect);
			sec = m_sections.Size()-1;
		}
		
		int val= FindVal(sec, key);
		if(val<0)
		{
			if(value.len())
			{
				m_changed=true;
				INIKEY ikey;
				ikey.name = key;
				ikey.val  = value;
				m_sections[sec].keys.Push(ikey);
			}
			return;
		}

		m_changed=true;
		if(value.len() == 0) // delete key
		{
			m_sections[sec].keys.Remove(val, true);
		}
		else
		{
			m_sections[sec].keys[val].val = value;
		}
	}

};



#endif // _GF_INI_H_INC_

