#ifndef NODE_H
#define NODE_H
/**
  @class NodeFactory
  @brief Factory for providing path-planner nodes.

  Copyright (C) 2003 Ted Cipicchio <ted@thereisnospork.com>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to
  deal in the Software without restriction, including without limitation the
  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  IN THE SOFTWARE.

  @date  Created: Thu Mar 20 2003
*/


//----------------------------------------------------------------------------
/// Generic path-planner node.
struct Node
{
	/// Parent node in path.
	Node *pParent;

	/// X-coordinate.
	int x;
	/// Y-coordinate.
	int y;

	/// Heuristic weight.
	float weight;

	/// Cost to travel up to this node from the start.
	float cost;

	/// Next node (used in hash table linked lists).
	Node *pNext;
};


//----------------------------------------------------------------------------
class NodeFactory
{
public:
	//------------------------------------------------------------------------
	/** @name Construction/Destruction */
	//@{

	/// Constructor.
	NodeFactory ();
	/// Destructor.
	virtual ~NodeFactory ();

	//@}


	//------------------------------------------------------------------------
	/** @name General Operations */
	//@{

	/// Allocate a pool of nodes.
	bool allocate (unsigned int size);

	/// Fetch a node from the pool.
	Node *fetch ();
	/// Release a node back into the pool.
	void release (Node *pNode);

	//@}

private:
	/// Size of node pool.
	unsigned int mv_nodePoolSize;
	/// Pool of nodes.
	Node **mv_nodePool;

	/// Index of top node in pool (next to give out).
	unsigned int mv_nodeTopIdx;
};


/// Fetch a node from the pool.
/**
  @return  A pointer to a node from the pool, or a null pointer if no more
           nodes are left.

  @note  Nodes fetched should ALWAYS be returned to the pool using the
         release() method.  This ensures they are properly deleted when the
         factory is deleted and no memory is leaked.

  @see release()
*/
inline Node *NodeFactory::fetch ()
{
	return (!mv_nodePool || mv_nodeTopIdx >= mv_nodePoolSize
		? 0
		: mv_nodePool[mv_nodeTopIdx++]);
}


/// Release a node back into the pool.
/**
  @param pNode  Node to return to the node pool.

  @note  Only nodes fetched from the factory using fetch() should be released.
         The node factory deletes nodes in its destructor based on what's in
         the pool at the time of destruction.  Putting nodes into the node
         pool that have not been created by the factory can cause memory
         leaks.

  @see fetch()
*/
inline void NodeFactory::release (Node *pNode)
{
	if (pNode && mv_nodeTopIdx)
//		mv_nodePool[--mv_nodeTopIdx] = pNode;
		--mv_nodeTopIdx;
}

#endif	// NODE_H
